import sys
import os
import warnings

from random import randint
from collections import OrderedDict

import numpy as np
import pytest

from mmgroup import MM0, MMSpace, MMV
from mmgroup.generators import gen_leech3_op_vector_word
from mmgroup.clifford12 import leech3matrix_echelon
from mmgroup.clifford12 import leech3matrix_compress
from mmgroup.clifford12 import xsp2co1_from_vect_mod3
from mmgroup.clifford12 import leech3matrix_sub_diag
from mmgroup.clifford12 import leech2matrix_add_eqn
from mmgroup.clifford12 import leech2matrix_solve_eqn
from mmgroup.clifford12 import leech3matrix_rank
from mmgroup.mat24 import MAT24_ORDER
from mmgroup.mm15 import op_word_tag_A
from mmgroup.mm15 import op_load_leech3matrix as mm_op15_load_leech3matrix
from mmgroup.mm3 import op_load_leech3matrix as mm_op3_load_leech3matrix

def rand_matrix_mod_n(n, dtype = np.int64):
    """Create random 24 x 24 matrix of intgers modulo n"""
    m =  np.zeros((24, 24), dtype = dtype)
    for i in range(24):
        for j in range(24):
            m[i,j] = randint(0,n-1)
    return m

def print_mod3(m):
    """Print 24 x 24 matrix of integers modulo 3"""
    m1 = m % 3
    for i in range(24):
        print("".join(map(str, m1[i] % 3)))
    print("")



def matrix_to_rep_modp(m, p):
    """Convert 24 x 24 matrix to monster rep, part A, modulo p

    The function returns the first block of a vector of the monster
    rep modulo p. Ths block corresponds to the entries of the
    vector with tag 'A'. 
    """
    assert p in (3, 15)
    if p == 3:
        m0 = (m - 1) % 3 + 1
        b = np.zeros(24, dtype = np.uint64)
        for i in range(24):
            b[i] = sum(m0[i,j] << (2*j) for j in range(24))
    elif p == 15:
        m0 = m % 3 + 3 * rand_matrix_mod_n(5)
        b = np.zeros(2*24, dtype = np.uint64)
        for i in range(24):
            r = sum((int(m[i,j]) % 3 ) << (4*j) for j in range(24))
            b[2*i] = r & 0xffffffffffffffff
            b[2*i + 1] = r >> 64
    else:
        err = "C functions do not support modulus p=%s"
        raise ValueError(err, str(p))
    return b



def from_array(m, load_mode = None):
    """Load 24 x 24 matrix 'm' to array of type 'leech3matrix'

    If load_mode is 3 or 15 then matrix is converted to a vector of
    of the monter rep using function ``matrix_to_rep_modp(m, p)``
    with ``p = load_mode``.

    By default we convert the matrix directly.

    The function returnes an numpy array that can be processed
    with the function in file ``leech3matrix.c``.
    """
    a = np.zeros(72, dtype = np.uint64)
    if load_mode == 15:
        b = matrix_to_rep_modp(m, load_mode)
        mm_op15_load_leech3matrix(b, a)
    elif load_mode == 3:
        b = matrix_to_rep_modp(m, load_mode)
        mm_op3_load_leech3matrix(b, a)
    elif not load_mode:
        for i in range(24):
            s = int(sum(int(m[i,j]) % 3 << 4*j for j in range(24)))
            a[3*i] = s & 0xffffffffffffffff
            a[3*i+1] = (s >> 64) & 0xffffffff
    else:
        err = "Loading from representation modulo % not suppoerted"
        raise ValueError(err % load_mode)
    return a


def as_array(a, high = 0, dtype = np.int64):
    """Convert  array of type 'leech3matrix' to 24 x 24 matrix"""
    m =  np.zeros((24, 24), dtype = dtype)
    jj = [96 * bool(high) + 4 * j for j in range(24)]
    jj = [(j >> 6, j & 0x3f) for j in jj]
    for i in range(24):
        for j, (ofs, shift) in enumerate(jj):
             m[i,j] = ((int(a[3*i + ofs]) >> shift) & 3) % 3
    return m
 
#######################################################################
# Test function leech3matrix_echelon
#######################################################################


@pytest.mark.qstate
def test_leech3matrix_echelon(verbose = 0):
    unit = np.eye(24)
    load_modes = [None, 15]
    for i in range(20):
        load_mode = load_modes[i % len(load_modes)]
        m = rand_matrix_mod_n(3, dtype = np.int64)
        a =  from_array(m, load_mode)
        leech3matrix_sub_diag(a, 2, 24)
        if verbose:
            print("load_mode =", load_mode)
            print_mod3(as_array(a,0))
            print_mod3(as_array(a,1))
        leech3matrix_echelon(a)
        if verbose:
            print_mod3(as_array(a,0))
            print_mod3(as_array(a,1))
        mi = as_array(a, high=1)
        prod = as_array(a, high=0)
        if verbose:
            print_mod3(prod)
            print_mod3(mi @ m)
            #print((mi @ m - prod) % 3)
        assert ((mi @ m - prod) % 3 == 0).all()




#######################################################################
# Test function leech3matrix_kernel_vector
#######################################################################

INT_BITS = 64

def kernel_testmatrix(p, tags = "dxypt"*3, diag = 0):
    space = MMV(p)
    a = space()
    diag %= 3
    for i in range(24):
        d = diag + randint(1, 2) + 3 * randint(0,4)
        a['A', i, i] = d % p if i else diag
    # Now ``a`` encodes the tag A part of the rep of the monster mod p
    # It encodes a symmetric matrix with kernel :math:`\Omega`.
    v3 = 1  # \Omega in Leech lattice mod 3 encoding
    g = MM0([(t,'r') for t in tags])
    a.ops.op_word_tag_A(a.data, g.mmdata, len(g.mmdata), 1)
    v3_kernel = gen_leech3_op_vector_word(v3, g.mmdata, len(g.mmdata))
    return a, v3_kernel

def neg_leech_mod3(x):
    t = (x ^ (x >> 24)) & 0xffffff
    return x ^ t ^ (t << 24)

def str_v3(v3):
    d = {0:0, 1:1, 0x1000000:2, 0x1000001:0}
    l = [str(d[(v3 >> i) & 0x1000001]) for i in range(24)]
    return "".join(l)


def load_leech3matrix(p, v):
    a = np.zeros(3*24, dtype = np.uint64)
    if p == 3:
        mm_op3_load_leech3matrix(v, a)
    elif p == 15:
        mm_op15_load_leech3matrix(v, a)
    else:
        err = "Computation of kernel vector not supporte for modulus %s"
        raise ValueError(err % p)
    return a


def kernel_vector(p, v, diag):
    a = load_leech3matrix(p, v)
    return leech3matrix_rank(a, diag) & 0xffffffffffff

@pytest.mark.qstate
def test_leech3matrix_kernel_vector(verbose = 0):
    print("Testing function leech3matrix_kernel_vector()")
    for i in range(10):
        for p in (3, 15):
            diag = randint(0,2)
            v_a, v3_kernel =  kernel_testmatrix(p, "pxydl" * 4, diag)
            a = v_a.data
            a1 = np.zeros(72, dtype = np.uint64)
            a2 = np.zeros(48, dtype = np.uint64)
            if verbose:
                print("p = %d, diagonal = %d" % (p, diag))
                print("Part of tag A matrix")
                print(v_a["A", :16, :16])
                print("Kernel vector mod 3 expected:", str_v3(v3_kernel))
            v3_obt = kernel_vector(p, a, diag)
            if verbose:
                print("Kernel vector mod 3 obtained:", str_v3(v3_obt))
                print("")
            a1 = load_leech3matrix(p, a)
            leech3matrix_sub_diag(a1, diag, 0);
            leech3matrix_sub_diag(a1, 2, 24);
            leech3matrix_echelon(a1);
            leech3matrix_compress(a1, a2)
            #for i in range(24): print(i, hex(a2[2*i]))
            #for i in range(24): print(i, hex(a2[2*i+1]))
            v3_obt1 = xsp2co1_from_vect_mod3(a2[47])
            hh = xsp2co1_from_vect_mod3
            #print(hex(v3_obt), hex(v3_obt1) ) 
            data = [v3_obt1, v3_kernel, neg_leech_mod3(v3_kernel)]
            assert v3_obt in data[1:], list(map(hex, data))
            assert v3_obt1 in data[1:], list(map(hex, data))
            


#######################################################################
# Test functions leech2matrix_add_eqn and leech2matrix_solve_eqn
#######################################################################
       

def one_test_leech2matrix_eqn(n, verbose = 0):
    ml = [1 << i for i in range(n)]
    for k in range(2 * n * n):
        i = randint(0, n-1)
        j = randint(0, n-1)
        if i != j:
            ml[i] ^= ml[j]
    if verbose:
        print("Test function leech2matrix_add_eqn")
        print("m =", [hex(x) for x in ml])
    m, rows = np.zeros(n, dtype = np.uint64), 0
    for i, l in enumerate(ml):
        found = leech2matrix_add_eqn(m, rows, n, l)
        assert found == 1
        rows += found
        if i == 2:
            l_bad = ml[0] ^ ml[1]
            assert leech2matrix_add_eqn(m, rows, n, l_bad) == 0
        if i == 10:
            l_bad = ml[0] ^ ml[3] ^ ml[6] ^ ml[9]
            assert leech2matrix_add_eqn(m, rows, n, l_bad) == 0
    v = randint(0, (1 << n) - 1) 
    m32 = np.array(m, dtype = np.uint32)
    w = leech2matrix_solve_eqn(m32, n, v)
    if verbose:
        print("inv =", [hex(int(x)) for x in m])
        print("v =", hex(v), ";", hex(w), " * m = v" )
    check = 0
    for i, l in enumerate(ml):
        if (w >> i) & 1:
            check ^=  l
    assert v == check


@pytest.mark.qstate
def test_leech2matrix_eqn():
    for n in range(2,33):
        one_test_leech2matrix_eqn(n)


       




import time
import numpy as np

from mmgroup import structures
from mmgroup.mat24 import vect_to_cocode
from mmgroup.mat24 import ploop_theta
from mmgroup.mm import mm_aux_index_sparse_to_leech2
from mmgroup.mm import mm_vector
from mmgroup.mm import mm_aux_mmv_extract_sparse_signs
from mmgroup.structures.mm0_group import MM0Group, MM0
from mmgroup.mm_group import MMGroup, MM
from mmgroup.mm_space import MMSpace, MMV
from mmgroup.generators import mm_group_check_word_n
from mmgroup.generators import mm_group_words_equ
from mmgroup.generators import mm_group_n_mul_element
from mmgroup.generators import mm_group_n_reduce_element 
from mmgroup.generators import gen_leech3to2_type4
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.clifford12 import chk_qstate12
from mmgroup.clifford12 import uint64_parity
from mmgroup.clifford12 import leech2matrix_add_eqn
from mmgroup.clifford12 import leech2matrix_solve_eqn
from mmgroup.clifford12 import bitmatrix64_t
from mmgroup.clifford12 import xsp2co1_half_order_word
from mmgroup.clifford12 import xsp2co1_power_word
from mmgroup.mm15 import op_eval_A_rank_mod3 as mm_op15_eval_A_rank_mod3 

from mmgroup.mm15 import op_copy as mm_op15_copy
from mmgroup.mm15 import op_compare as mm_op15_compare
from mmgroup.mm15 import op_word as mm_op15_word
from mmgroup.mm15 import op_norm_A as mm_op15_norm_A
from mmgroup.mm15 import op_watermark_A as mm_op15_watermark_A
from mmgroup.mm_reduce import mm_order_element_M
from mmgroup.mm_reduce import mm_order_element_Gx0
from mmgroup.mm_reduce import mm_order_load_vector
from mmgroup.mm_reduce import mm_order_load_tag_data
from mmgroup.mm_reduce import mm_reduce_M



MMV3 = MMV(3)
MMV15 = MMV(15)
MM = MM0  #  TODO: Fixme




def get_order_vector(*args, **kwds):
    v = mm_vector(15)
    mm_order_load_vector(v.data)
    return v

def get_order_tag_vector(*args, **kwds):
    a = np.zeros(97, dtype = np.uint32)
    res = mm_order_load_tag_data(0, a, len(a))
    assert res >= 0
    return a[:res]


###########################################################################
# Check equality of two elements of the monster
###########################################################################
 

def check_mm_equal(g1, g2, mode = 0):
    """Return ``g1 == g2`` for elements ``g1, g2`` of the monster.

    If ``mode == 0`` (default) we first try to check equality inside 
    in the subgroup ``N_0`` of the monster, which may be considerbly 
    faster. 

    If ``mode != 0`` or this is not possible we check if 
    ``v * g1 * g2**(-1) == v`` holds for the *ORDER_VECTOR* ``v``. 

    We just check the data in ``g1`` and ``g2``, ingnoring
    ``g1.group`` and ``g2.group``.
    """
    assert isinstance(g1, (MM, MM0))
    assert isinstance(g2, (MM, MM0))
    g3 = np.zeros(2 * (g1.length + g2.length) + 1, dtype = np.uint32)
    status = mm_group_words_equ(g1._data, g1.length,
        g2._data, g2.length, g3)
    if status < 2:
        return not status
    v = mm_vector(15)
    mm_order_load_vector(v.data)
    work = mm_vector(15)
    mm_op15_word(v, g3, status - 2, 1, work)
    mm_order_load_vector(work.data)
    return not mm_op15_compare(v, work)



###########################################################################
# Computing the order of an element of the monster
###########################################################################






def check_mm_order(g, max_order = 119):
    """Return order of monster group element ``g``.

    ``g`` must be an instance of class ``MM``.  The function
    returns the order of ``g`` if ``max_order`` is set to its default
    value. 

    Computing the order of an element of the monster is time consuming; 
    and in some cases we are interested in small orders only.   
    If ``max_order``is given then the function may return 0 if the
    order of ``g`` is greater than ``max_order``.
    """
    assert isinstance(g, (MM0, MM))
    g.reduce()
    o = mm_order_element_M(g._data, g.length, max_order)
    return  chk_qstate12(o)

def check_mm_half_order(g, max_order = 119):
    """Return (halved) order of monster group element ``g``.

    ``g`` must be an instance of class ``MM``.  The function
    returns a pair ``(o, h)`` where ``o`` is the order of ``g``, and
    ``h = g**(o/2)`` for an even ``o``. We put ``h = None`` if
    ``o`` is odd.

    Parameter ``max_order`` is as in function ``check_mm_order``. 

    If ``h`` is in the subgroup :math:`G_{x0}`` then ``h`` is 
    returned as a word in the generators of that subgroup.
    """
    assert isinstance(g, (MM0, MM))
    g.reduce()
    h = np.zeros(10, dtype = np.uint32)
    o1 = mm_order_element_Gx0(g._data, g.length, h, max_order)
    chk_qstate12(o1)
    if o1 == 0:
        return 0, None
    h = h[:o1 & 0xff]
    o1 >>= 8
    h2 = np.zeros(10, dtype = np.uint32)
    o2 = xsp2co1_half_order_word(h, len(h), h2)
    h2 = h2[:o2 & 0xff]
    o2 >>= 8
    o = o1 * o2
    if (o2 & 1) == 0:
        return o, g.group('a', h2).reduce()
    if (o & 1):
        return o, None
    if o2 == 1:
        return o, g ** (o1 >> 1)
    # compute q, r, such that the result is  (g**o1)**q  * g**r
    q, r = divmod(o >> 1, o1) 
    w = g.group('a', h)
    return o, (w**q * g**r).reduce()

###########################################################################
# Check if an element of the monster is in the subgroup G_x0
###########################################################################
 




def check_mm_in_g_x0(g):
    """Check if ``g`` is in the subgroup ``G_x0`` of the monster
   
    If ``g`` is in the subgroup ``G_x0`` of the monster then the
    function changes the word representing ``g`` to a (uniquely 
    defined) word in the generators of the subgroup ``G_x0`` and
    returns ``g``.  
    
    Otherwise the function does not change ``g`` and returns ``None``.
    
    ``g`` must be an instance of class 
    ``mmgroup.mm_group.MM``.
    """
    g1 = np.zeros(10, dtype = np.uint32)
 
    res = chk_qstate12(mm_order_element_Gx0(g._data,  g.length, g1, 1))
    #print("RES", hex(res))
    if ((res >> 8) != 1):
        return None 
    length = res & 0xff
    assert length <= 10
    g._extend(10)
    g._data[:length] = g1[:length]
    g.length = length
    g.reduced = 0
    g.reduce()
    return g








r"""

The functions in file ``gen_leech.c`` implement the operation of the
subgroup :math:`G_{x0}` (of structure :math:`2^{1+24}.\mbox{Co}_1`)
of the monster group on its extraspecial subgroup :math:`Q_{x0}` 
(of structure :math:`2^{1+24}`) by conjugation.

Here an element of the group :math:`Q_{x0}` is given as a 25-bit 
integer in Leech lattice encoding. From that encoding we obtain an 
encoding of :math:`\Lambda / 2 \Lambda` (where :math:`\Lambda` is 
the Leech lattice) by dropping the most significant bit. This 
corresponds to the isomorphism
:math:`Q_{x0} / \{\pm1\} \cong \Lambda / 2 \Lambda`.

An element of the group  :math:`G_{x0}` is encoded as an array of 
32-bit integers, where each integer corresponds to a generator of
the group, as described in the documentation of the header file
``mmgroup_generators.h``.
  
A vector in the Leech lattice modulo 2 has a type and also a subtype
as described in **The mmgroup guide for developers**, section
:ref:`computation-leech2`. Here the subtype is a two-digit decimal
number, where the first digit is the type. Function 
``gen_leech2_subtype`` returns the subtype as a BCD-coded integer.
E.g. the subtype 46 is returned as the hexadecimal integer 0x46. 
So the type can be obtained from the subtype via a shift operation.

For computations in the group :math:`G_{x0}` or :math:`\mbox{Co}_1`
it is important to find an element of :math:`G_{x0}` that maps an 
arbitrary type-4 vector in :math:`\Lambda / 2 \Lambda` to the unique
type-4 vector :math:`\Omega`, see :ref:`computation-leech2`.
Function ``gen_leech2_reduce_type4`` performs this task.

The construction of the group :math:`G_{x0}` also requires some
computations in the automorphism group :math:`\mbox{Co}_0`
(of structure :math:`2.\mbox{Co}_1`) of  :math:`\Lambda`, see
:cite:`Con85`, section 9 or :cite:`Seysen20`, section 9. The group
:math:`\mbox{Co}_0` acts faithfully on :math:`\Lambda / 3 \Lambda`. 
Therefore file ``gen_leech.c`` also provides some functions for 
dealing with vectors in the Leech lattice modulo 3.

For any :math:`w \in \Lambda` the vector 
:math:`v = \sqrt{8} \cdot w` has integral coordinates. We 
represent a vector :math:`w \in \Lambda / 3 \Lambda` by the vector  
:math:`v = \sqrt{8} \cdot w`, with the coordinates of :math:`v` 
taken modulo 3. We represent each coordinate 
:math:`v_i, 0 \leq i < 24` as a two-bit integer;
so all 48 bits of :math:`v` fit into the a 64-bit integer. Let
:math:`v_{i,1}, v_{i,0}` be the bits of :math:`v_{i}` of 
valence :math:`2^1` and :math:`2^0`, respectively. Then we
encode :math:`v` as an 48-bit integer :math:`x` so that bit 
:math:`24 \cdot j+i` of :math:`x` is equal to :math:`v_{i,j}`. 
This encoding looks peculiar, but it greatly simplifies the 
interaction with the functions in file ``mat24_functions.c``. We 
call this encoding the **Leech lattice mod 3** encoding.
On input, both bits :math:`v_{i,1}, v_{i,0}` may be equal to 1;
but on output, at most on of these bits is equal to 1. 

In :cite:`Seysen20`, section 9.3, the generators  
:math:`x_d, x_\delta, y_\delta, x_\pi, \xi` of :math:`G_{x0}`
are also defined as generators of a group  :math:`G_{x1}` with 
:math:`|G_{x1}:G_{x0}| = 2`. The group :math:`G_{x1}` operates
on :math:`\Lambda`, but the group :math:`G_{x0}` does not.
So we take these generators as generators of :math:`G_{x1}`.
The kernel of the operation of :math:`G_{x1}` on :math:`\Lambda`
is the group  :math:`Q_{x0}` generated by the elements
:math:`x_d, x_\delta, y_{-1}, d \in \mathcal{P}, 
\delta \in  \mathcal{C}`. Here the element :math:`y_{\Omega}`
acts as the central element :math:`-1`.

"""



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import collections
import re
import warnings
from numbers import Integral
import numpy as np


from mmgroup.bitfunctions import isinteger, v2, hibit

from mmgroup.generate_c import c_snippet, TableGenerator, make_table
from mmgroup.generate_c import UserDirective, UserFormat


from mmgroup.dev.mm_basics.mm_tables_xi import MM_TablesXi
from mmgroup.dev.mm_op.mm_op import MM_Op


class MonomialOpTableInfo:
    def __init__(self, shape, start, op_diff):
        self.SHAPE = shape
        self.START = start 
        self.OP_DIFF = op_diff


class MonomialOp_xi_uint8_t(MM_Op):
    """Standard implementation for class MonomialOp_xi

    This is the standard implementation for the monomial cases
    of operation xi.
      
    In this case the main loop is:

    uint8_t b[2496], *p_b;
    for (i = 0; i < {i_src.SHAPE[0]}; ++i) {
        p_b = b;
        
        for (j = 0; j < {i_src.SHAPE[1]}; ++j) {
           // %%OP_XI_LOAD p_src, p_b, {i_src.SHAPE[2]}
           p_src += {V24_INTS};
           p_b += 32;
        }
        
        for (j = 0; j < {i_dest.SHAPE[1]}; ++j) {
           // %%OP_XI_STORE b, p_perm, p_sign, p_dest, {i_dest.SHAPE[2]}
           p_dest += {V24_INTS};
           p_perm += {i_dest.SHAPE[2]};
           p_sign += 1;
        }        
    }

    """

    def __init__(self, p):
        super(MonomialOp_xi_uint8_t, self).__init__(p)
        MM_TablesXi()
        self.make_table_info()
        # make tables and directives for code generation
        self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())

    def load_bytes(self, source, bytes, n_bytes, type_=""):
        s = ""
        n_registers = 8 // self.FIELD_BITS
        registers =  ", ".join(["r%d" % i for i in range(n_registers)])
        s = "uint_mmv_t %s;\n"% registers
        LOAD_MASK = self.hex(self.smask(self.P, -1, 8))
        type_ = "(%s)" % type_ if len(type_) else ""
        for i, start in enumerate(range(0, n_bytes, self.INT_FIELDS)):
            s += "r0 =  (%s)[%d];\n" % (source, i)
            for j in range(1, n_registers):
                s += "r%d = (r0 >> %d) & %s;\n" % (
                   j, j * self.FIELD_BITS, LOAD_MASK)
            s += "r0 &= %s;\n" % LOAD_MASK
            for j in range(self.INT_FIELDS):
                if start + j < n_bytes:
                    s += "(%s)[%d] = %s(r%d >> %d);\n" % (bytes,  start + j, 
                        type_, j % n_registers, 8 * (j // n_registers))
        return s;



    def store_bytes(self, bytes, perm, sign, dest, n_bytes):
        s = "uint_mmv_t r0, r1;\n"

        d = ["((uint_mmv_t)(%s[%s[%d]]) << %d)" % (bytes, perm, i, 
                  self.FIELD_BITS * (i % self.INT_FIELDS)) 
                  for i in range(n_bytes)
            ]
        for i, start in enumerate(range(0, n_bytes, self.INT_FIELDS)):
            s += "r0 = " 
            s += "\n  + ".join(d[start:start+self.INT_FIELDS])
            s += ";\n" 
            s += "r1 = (%s)[0] >> %s;\n" % (sign, start)
            s += self.gen_mmv_uint_spread("r1", "r1")
            s += "(%s)[%d] = r0 ^ r1;\n" % (dest, i)

        #s +=("// i=%d\n" % i)
        for j in range(i+1, 32 // self.INT_FIELDS):
            s += "%s[%d] = 0;\n" % (dest,j)
        return s


    def comment(self,i):
        src_t = MM_TablesXi.SOURCE_TAGS[i]
        dest_t = MM_TablesXi.DEST_TAGS[i]
        if (src_t[0] == src_t[1] and dest_t[0] == dest_t[1]):
            return "// Map tag %s to tag %s."% (src_t[0], dest_t[0])
        s = "// Map tag %s to tag %s if e = 1\n" % (src_t[0], dest_t[0])
        s += "// Map tag %s to tag %s if e = 2" % (src_t[1], dest_t[1])
        return s


    def make_table_info(self):
        self.table_info = []
        sh = self.LOG_INT_FIELDS
        for i in range(1,6):
            c = self.comment(i)
            info_source =  MonomialOpTableInfo(
                MM_TablesXi.SOURCE_SHAPES[i],
                MM_TablesXi.SOURCE_START_1[i] >> sh, 
                MM_TablesXi.SOURCE_OP_DIFF[i]
            )
            info_dest =  MonomialOpTableInfo(
                MM_TablesXi.DEST_SHAPES[i],
                MM_TablesXi.DEST_START_1[i] >> sh, 
                MM_TablesXi.DEST_OP_DIFF[i]
            )
            self.table_info.append((i-1, info_source, info_dest, c))    
        self.table_diff = MM_TablesXi.MAX_ABS_START_DIFF >> sh

    def make_tables(self):
        return {
            "OP_XI_TABLE_INFO": self.table_info, 
            "OP_XI_TABLE_DIFF": self.table_diff
        }
     
    def make_directives(self):
        return {
            "OP_XI_LOAD": UserDirective(self.load_bytes, "ssis"),
            "OP_XI_STORE": UserDirective(self.store_bytes, "ssssi"),
        }



class MonomialOp_xi_uint_fast8_t(MonomialOp_xi_uint8_t):
    """This is a variant of class MonomialOp_xi_uint8_t

    Here the main loop is as in class ``MonomialOp_xi_uint8_t``.
    But the variable ``b`` may also be integer type.
 
    It might save a *partial register stall* when storing data to a
    variable of type ``uint_fast8_t`` instead of type ``uint8_t``
    """
    n_regs = {1:1, 2:2, 4:2, 8:2, 16:4, 32:4, 64:8}

    def __init__(self, p):
        super(MonomialOp_xi_uint_fast8_t, self).__init__(p)
        self.n_registers = self.n_regs[self.INT_FIELDS]

    def load_bytes(self, source, bytes, n_bytes):
        s = ""
        n_registers = self.n_registers
        registers =  ", ".join(["r%d" % i for i in range(n_registers)])
        s = "uint_mmv_t %s;\n"% registers
        WIDTH = n_registers * self.FIELD_BITS
        LOAD_MASK = self.hex(self.smask(self.P, -1, WIDTH))
        for i, start in enumerate(range(0, n_bytes, self.INT_FIELDS)):
            s += "r0 =  (%s)[%d];\n" % (source, i)
            for j in range(1, n_registers):
                s += "r%d = (r0 >> %d) & %s;\n" % (
                   j, j * self.FIELD_BITS, LOAD_MASK)
            s += "r0 &= %s;\n" % LOAD_MASK
            for j in range(self.INT_FIELDS):
                if start + j < n_bytes:
                    s += "(%s)[%d] = r%d >> %d;\n" % (bytes,  start + j, 
                        j % n_registers, WIDTH * (j // n_registers))
        return s;

    def store_bytes(self, bytes, perm, sign, dest, n_bytes):
        s = "uint_mmv_t r0, r1;\n"

        d = ["((uint_mmv_t)(%s[%s[%d]]) << %d)" % (bytes, perm, i, 
                  self.FIELD_BITS * (i % self.INT_FIELDS)) 
                  for i in range(n_bytes)
            ]
        n_registers = self.n_registers

        for i, start in enumerate(range(0, n_bytes, self.INT_FIELDS)):
            reg = "r0 = ("
            d1 = d[start:start+self.INT_FIELDS]
            M = self.smask(self.P, range(n_registers))
            while len(d1):
                hd, d1 = d1[:n_registers], d1[n_registers:]
                s += reg + "\n  + ".join(hd) + ") & %s;\n" % self.hex(M)
                reg = "r0 += ("
                M <<= n_registers * self.FIELD_BITS

            s += "r1 = (%s)[0] >> %s;\n" % (sign, start)
            s += self.gen_mmv_uint_spread("r1", "r1")
            s += "(%s)[%d] = r0 ^ r1;\n" % (dest, i)

        #s +=("// i=%d\n" % i)
        for j in range(i+1, 32 // self.INT_FIELDS):
            s += "%s[%d] = 0;\n" % (dest,j)
        return s



class MonomialOp_xi_alternative(MM_Op):
    """Alternative implementation to class MonomialOp_xi
 
    In this case the main loop is:

    for (i = 0; i < {i_dest.SHAPE[0]}; ++i) {
        for (j = 0; j < {i_dest.SHAPE[1]}; ++j) {
           // %%OP_XI p_src, p_perm, p_sign, p_dest, {i_dest.SHAPE[2]}
           p_dest += {V24_INTS};
           p_perm += {i_dest.SHAPE[2]};
           p_sign += 1;
        }
        p_src += {int:i_src.SHAPE[1] * V24_INTS};
    }

    Although here fewer data are being moved, this leads to a slower
    implementation on a 64 bit 80x86 CPU.
    """
    def __init__(self, p):
        super(MonomialOp_xi, self).__init__(p)
        self.make_table_info()
        self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())

    def make_table_info(self):
        self.table_info = []
        sh = self.LOG_INT_FIELDS
        for i in range(1,6):
            c = self.comment(i)
            info_source =  MonomialOpTableInfo(
                MM_TablesXi.SOURCE_SHAPES[i],
                MM_TablesXi.SOURCE_START_1[i] >> sh, 
                MM_TablesXi.SOURCE_OP_DIFF[i]
            )
            info_dest =  MonomialOpTableInfo(
                MM_TablesXi.DEST_SHAPES[i],
                MM_TablesXi.DEST_START_1[i] >> sh, 
                MM_TablesXi.DEST_OP_DIFF[i]
            )
            self.table_info.append((i-1, info_source, info_dest, c))    
        self.table_diff = MM_TablesXi.MAX_ABS_START_DIFF >> sh

    def comment(self,i):
        src_t = MM_TablesXi.SOURCE_TAGS[i]
        dest_t = MM_TablesXi.DEST_TAGS[i]
        if (src_t[0] == src_t[1] and dest_t[0] == dest_t[1]):
            return "// Map tag %s to tag %s."% (src_t[0], dest_t[0])
        s = "// Map tag %s to tag %s if e = 1\n" % (src_t[0], dest_t[0])
        s += "// Map tag %s to tag %s if e = 2" % (src_t[1], dest_t[1])
        return s


    def op(self, src, perm, sign, dest, length):
        assert length in (24, 32)
        INT_FIELDS = self.INT_FIELDS
        s = "uint_mmv_t r0, rd, rp;\n"
        for i in range(0, length, INT_FIELDS): 
            op = ""
            for j in range(0, min(INT_FIELDS, length - i)):
                s += "rp = (%s)[%d];\n" % (perm, i + j)
                s += "rd = (%s)[rp >> %d] >> ((rp & %d) << %d);\n" % (
                      src, self.LOG_INT_FIELDS, INT_FIELDS - 1, 
                      self.LOG_FIELD_BITS
                )
                s += "r0 %s= ((rd & %d) << %s);\n" % (
                     op, self.p, j * self.FIELD_BITS
                )
                op = "+"
            s += "rd = (%s)[0] >> %s;\n" % (sign, i)
            s += self.gen_mmv_uint_spread("rd", "rd")
            s += "(%s)[%d] = r0 ^ rd;\n" % (dest, i // INT_FIELDS)
        for i in range(1 + (length - 1) // INT_FIELDS, 32 // INT_FIELDS):
            s += "(%s)[%d] = 0;\n"   % (dest, i)
        return s    

    def make_tables(self):
        return {
            "OP_XI_TABLE_INFO": self.table_info, 
            "OP_XI_TABLE_DIFF": self.table_diff
        }
     
    def make_directives(self):
        return {
            "OP_XI": UserDirective(self.op, "ssssi"),
        }


 

MonomialOp_xi =  MonomialOp_xi_uint8_t

class Mockup_MonomialOp_xi:
    def __init__(srlf, *args):
        pass
    tables =  {
        "OP_XI_TABLE_INFO": [], 
        "OP_XI_TABLE_DIFF": 0
    }
    def op(self, *args):
        return "\n"
    directives = {
        "OP_XI": UserDirective(op, "ssssi")
    }
        
        


import numpy as np
import datetime
import time

from mmgroup import MMGroup, MM0, MM
from mmgroup.structures.mm_order import check_mm_order
from mmgroup.structures.xsp2_co1 import Xsp2_Co1



def mm_conjugate_involution(g, check=True, ntrials=20, verbose = 0):
    """Conjugte an involution in the monster to a standard element

    Given an involution :math:`g` in the monster group, the function 
    tries to find an element :math:`h` with :math:`h^{-1} g h = z`,
    where :math:`z` a standard representative of the involution class
    as in method ``conjugate_involution`` of class ``Xsp2_Co1``.
    """
    if not isinstance(g, (MM,MM0)):
        err = "Object must be element of the monster of type MM"
        raise TypeError(err)
    m = g.group
    m_1, z = m(), m('x', 0x1000)
    in_G_x = g.in_G_x0()
    if (verbose or check) and g*g != m_1:
        err = "Element is not an involution in the monster group"
        raise ValueError(err)
    if in_G_x:
        return Xsp2_Co1(g).conjugate_involution(m)
    if verbose:
        print("Function mm_conjugate_involution: conjugate g to the centre")
        print("g=", g)
    for i in range(ntrials):
        if verbose:
             print("\nmm_conjugate_involution trial", i)
        s = m('r', 1 + max(3, (i >> 2))) if i else m_1
        x = g**s
        o, y = (x * z).half_order(60)
        if verbose:
           print("x = g**s; s =", s)
           print("x * z has order o =", (x * z).order())
        if o == 0 or o & 1 or y is None:
            continue
        # Here o is even and y = (x * z)**(o/2) is an involution,
        # and y commutes with x and with z. Thus y is in G_x0.
        # Next we represent y as an element of G_x0.
        assert y.in_G_x0()
        if verbose:
            assert (x*z)**(o>>1) == y
            assert y**2 == m_1
            print("y = (x*z)**(o/2) = ", y)
            print("y has characters",  y.chi_G_x0())
        # Try to conjugate y to the central element z.
        # Continue the loop if this fails.
        try:
            itype = 3
            itype, h1 = Xsp2_Co1(y).conjugate_involution(m)
            assert itype == 2
        except (ValueError, AssertionError):
            if verbose:
                print("itype should be 2 but found %d" % itype) 
            continue
        # Now y**h1 = z.
        if verbose:
            print("itype=", itype)
            print("h1=", h1)
            y_conj = y**h1
            y_conj.in_G_x0()
            assert y_conj == z, y_conj
        x1 = x**h1
        # x1 commutes with y**h1 = z; so x1 is in G_x0.
        # Next we represent x1 as an element of G_x0
        x1.in_G_x0()
        if verbose:
            print("x1=", x1)
            print("x1 has characters",  x1.chi_G_x0())
        # Try to conjugate x1 to the central element z.
        # Raise ValueError the loop if this fails.
        itype, h2 = Xsp2_Co1(x1).conjugate_involution(m)
        # Now x1**h2 = x**(h1*h2) = g**(s*h1*h2) = z.
        # So we may return s*h1*h2
        t = (s*h1*h2).reduce()
        t.in_G_x0()
        t.reduce()
        if verbose:
            print("Result found (itype = %s):" % itype)
            print(t)
            print("Function mm_conjugate_involution terminated successfully.\n")
        return itype, t
    err = "Conjugation of element to central involution failed"
    raise ValueError(err)


def reduce_via_power(g, ntrials=20, verbose = 0):
    if verbose:
        print("\nTrying to shorten the element g of the monster:")
        print("g =", g)
    for i in range(ntrials):
        x0 = g.group("r", max(1, (i >> 2)))
        #x0 = g.group("r", "G_x0")
        g1 = g * x0
        o, g2 = g1.half_order()
        if o & 1 or g2 is None:
            continue
        try:
            it, h = mm_conjugate_involution(g2, check=False, ntrials=1, 
                      verbose = 0)
            assert it == 2
        except (ValueError, AssertionError):
            if verbose:
                err = "Conjugation of element to central involution failed"
                print(err + "\n")
            continue
        # Now g2**h = z, so g1**h  is in G_x0
        g1h = g1**h
        g1h.in_G_x0()
        g1_new = g1h**(h**-1)
        if verbose:
            assert g1_new == g1
        g_new = g1_new * x0**-1
        g_new.reduce()
        if verbose:
            print("Shortened value of element g:")
            print(g_new)
            assert g_new == g
            s = "Element of the monster has been shortened successfully"
            print(s + "\n")
        return g_new
    err = "Shortening of an element of the monster failed"
    raise ValueError(err)

        






from random import randint  
from functools import reduce
from operator import __or__, __xor__
import time

import numpy as np
import pytest

from mmgroup import MM0
from mmgroup import mat24
from mmgroup.mat24 import MAT24_ORDER 
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.structures.xsp2_co1 import Xsp2_Co1
from mmgroup.clifford12 import chk_qstate12
from mmgroup.clifford12 import xsp2co1_mul_elem_word
from mmgroup.clifford12 import xsp2co1_elem_monomial_to_xsp
from mmgroup.clifford12 import xsp2co1_xspecial_vector
from mmgroup.clifford12 import xsp2co1_elem_to_word
from mmgroup.clifford12 import xsp2co1_reduce_word
from mmgroup.clifford12 import xsp2co1_order_elem
from mmgroup.clifford12 import xsp2co1_half_order_word
from mmgroup.clifford12 import xsp2co1_order_word
from mmgroup.clifford12 import xsp2co1_power_elem



#######################################################################
# Function monomial_to_word
#######################################################################


def print_monomial_4096(monomial, explain = False):
    print("Monomial operation of element (up to sign):") 
    if explain:
        print("Unit vector 1 << i is mapped (row 0) XOR (row i)")
    for i in range(13):
        bits = [(monomial[i] >> j) & 1 for j in range(11,-1,-1)]
        print("  " + "".join(map(str, bits)))

def monomial_to_word(elem, verbose = 0):
    assert isinstance(elem, Xsp2_Co1)
    assert isinstance(verbose, int) , verbose   
    if verbose:
        print("Convert monomial element g of G_x0 to word. g is:")
        print(elem)
    a = []
    qs_i = elem.qs_t
    monomial = qs_i.monomial_row_op()
    if verbose:
        print_monomial_4096(monomial, explain = True)
    y = (monomial[12] & 0x7ff)
    mat24.matrix_from_mod_omega(monomial[1:])
    perm = mat24.autpl_to_perm(monomial[1:])
    perm = mat24.inv_perm(perm)
    pi = mat24.perm_to_m24num(perm)
    if pi:
        a.append(pi + 0xA0000000)
    if y:
        a.append(y + 0xC0000000)
    a = np.array(a, dtype = np.uint32)
    if verbose:
        print("Multiplier is:", MM.from_data(a))
        print_monomial_4096(monomial, explain = True)
    return a


def monomial_to_word_C(elem):
    a = np.zeros(2, dtype = np.uint32)
    elem_data = np.array(elem.data, dtype = np.uint64)
    len_a = xsp2co1_elem_monomial_to_xsp(elem_data, a)
    return a[:len_a]


#######################################################################
# Function elem_to_word
#######################################################################



def elem_to_word(elem, verbose = 1):
    assert isinstance(elem, Xsp2_Co1)
    assert isinstance(verbose, int) , verbose   
    group = elem.group
    len_a = 0
    a0 = np.zeros(16, dtype = np.uint32)
    img_Omega = elem.xsp_conjugate(0x800000)

    if verbose:
        print("Convert element g of G_x0 to word. Element is:")
        print(elem)
        print("g conjugates 0x800000 to %s" % hex(img_Omega)) 

    len_a =  gen_leech2_reduce_type4(img_Omega, a0)
    a0 = list(a0[:len_a])
    elem_reduced = elem.copy().mul_data(a0)

    if verbose:
        print("Word w0 stabilizing Omega:")
        print(" ".join([hex(x) for x in a0]))
        print("After multiplication with w0, the element is:")
        print(elem_reduced)
    
    a1 = list(monomial_to_word(elem_reduced, verbose > 1))
    elem_reduced = elem_reduced.mul_data(a1)

    x = elem_reduced.as_xsp()
    x ^= mat24.ploop_theta(x >> 12)
    a = []
    if (x & 0x1fff000):
        a.append(0x30000000 + (x >> 12))
    if (x & 0xfff):
        a.append(0x10000000 + (x & 0xfff))
    a +=  [x ^ 0x80000000 for x in (a0 + a1)[::-1]]
    a = np.array(a, dtype = np.uint32)
    if verbose:
        print("Word is:\n " + str(MM.from_data(a)))
    return a
 

def elem_to_word_C(elem):
    a = np.zeros(10, dtype = np.uint32)
    elem_data = np.array(elem.data, dtype = np.uint64)
    len_a = chk_qstate12(xsp2co1_elem_to_word(elem_data, a))
    return a[:len_a]

def reduce_word_C(w):
    a = np.array(w, dtype = np.uint32)
    a1 = np.zeros(10, dtype = np.uint32)
    len_a1 = chk_qstate12(xsp2co1_reduce_word(a, len(a), a1))
    return a1[:len_a1]
    


   
#######################################################################
# Yield test vectors as list of generator tuples
#######################################################################

from mmgroup.mat24 import MAT24_ORDER


MAX_TAG_ENTRY = {
   'd':0xfff, 'x':0x1fff, 'y':0x1fff, 'l':2, 'p': MAT24_ORDER-1
}
                 
                

def rand_G_x0_testword(t):
    """Yield element of G_{x0} as a list of pairs (tag, data) 

    ``t`` is a string of characters of generator tags, each character 
    must be in "dxypl"
    """
    w = []
    for tag in t:
         w.append((tag, randint(1, MAX_TAG_ENTRY[tag])))
    return w



def make_testwords(ntests = 100, monomial = True):
    test_words = [
       [],
       [('x', 0x800)],
       [('x', 0x1), ('y', 0x1), ('d', 0x800)],
       [('d', 0x800), ('y', 0x1)],
       [('d', 0x436), ('y', 0x12)],
       [('d', 0x837), ('y', 0x12)],
       [('x', 0x436), ('p', 0)],
       [('x', 0x436), ('p', 33445)],
       [('p',1234)],
    ]
    for w in test_words:
        yield w 
    tag_sets = ['d', 'dx', 'dxp', 'dxpy', 'dxpyl']
    for i in range(1,5):
         tag_sets.append('dxp' + 'lp' * i)
    for n in range(ntests):
        for t in tag_sets:
            if not monomial or not 'l' in t:
                 yield  rand_G_x0_testword(t)


#######################################################################
# Test function monomial_to_word
#######################################################################

@pytest.mark.xsp2co1
def test_monomial_to_word(ntests = 10, verbose = 0):
    print("Test function test_monomial_to_word()")
    for i, w in enumerate(make_testwords()):
        if verbose:
            print("\nTest %d:" % (i+1))
            m = MM0(w)
            print("Testing word\n%s" % m)
        elem = Xsp2_Co1(w)
        if verbose:
            print("This is element\n%s" % elem)
        data_i = monomial_to_word(elem, verbose > 1)
        elem_0 = elem.copy().mul_data(data_i)
        try:
            x = elem_0.as_xsp()
            ok = True
        except:
            raise
            ok = False
        if verbose or not ok:
            m_i = MM0("a", data_i)
            print("Reduced inverse word\n%s" % str(m_i))
            if not ok:
                print("Product with inverse\n%s" % elem_0)
                err = "WTF"
                raise ValueError(err)
            else:
                print("Result is:", hex(x))
            print("Test %d ok" % (i+1))
        data_C = monomial_to_word_C(elem)
        assert (data_C == data_i).all(), (data_C, data_i)


#######################################################################
# Test function elem_to_word
#######################################################################

@pytest.mark.xsp2co1
def test_elem_to_word(ntests = 50, verbose = 0):
    print("Test function test_elem_to_word()")
    for i, w in enumerate(make_testwords(monomial=False, ntests=ntests)):
        m = MM0(w)
        if verbose:
            print("\nTest %d:" % (i+1))
            print("Testing word\n%s" % m)
        elem = Xsp2_Co1(w)
        if i < 100:
            assert m == MM0(elem)
        if verbose:
            print("This is element\n%s" % elem)
        word = elem_to_word(elem, verbose > 1)
        if verbose:
            print("Reduced word\n%s" % MM0('a', word))
        elem_1 = Xsp2_Co1('a', word)
        if verbose:
            print("Recomputed element\n%s" % elem_1)
        ok = (elem == elem_1).all()
        if verbose or not ok:
            if not ok:
                print("Instead of to 1, the element reduces to:")
                print(elem * elem_1**(-1))
                err = "WTF"
                raise ValueError(err)
        word_C = elem_to_word_C(elem)
        assert (word_C == word).all(), (word_C, word)
        word1_C = reduce_word_C(m.mmdata)
        assert (word1_C == word).all(), (word1_C, word)
            


#######################################################################
# Test computing the power of an element of G_x0
#######################################################################

def xsp2co1_ref_power(wx, e):
    """Foolproof exponentiation in G_x0"""
    assert isinstance(wx, Xsp2_Co1)
    if e > 1:
        h = xsp2co1_ref_power(wx, e >> 1)  
        return h * h * wx if e & 1 else h * h
    if e == 1:
        return wx
    if e == 0:
        return Xsp2_Co1()
    return (wx**(-1)) ** (-e)


def xsp2co1_fast_power(wx, e):
    """The safe exponentiation in G_x0 to be tested"""
    assert isinstance(wx, Xsp2_Co1)
    power = Xsp2_Co1()
    chk_qstate12(xsp2co1_power_elem(wx._data, e, power._data))
    return power

@pytest.mark.xsp2co1
def test_elem_power(ntests = 50, verbose = 0):
    print("Test function computation a of power in G_x0")
    for i, w in enumerate(make_testwords(monomial=False, ntests=ntests)):
        e = randint(-2**35, 2**35)
        #print("Test ", i, "e", e)
        wx = Xsp2_Co1(w)
        power =  xsp2co1_fast_power(wx, e)
        ref_power = xsp2co1_ref_power(wx, e) # wx ** e
        assert power == ref_power, e
        if i < 10:
            e1 = i - 5
            power =  xsp2co1_fast_power(wx, e1)
            assert power == xsp2co1_ref_power(wx, e1) 


#######################################################################
# Test computing the order of an element of G_x0
#######################################################################


def xsp2co1_ref_order(wx):
    assert isinstance(wx, Xsp2_Co1)
    o = wx.qs.order(120)
    if o & 1 == 0:
        o = o >> 1
    unit, pw = wx.group(), wx**o
    if pw == unit:
         return o
    for i in range(2):
        o, pw = 2*o, pw * pw
        if pw == unit:
            return o
    err = "Order of QStateMatrix object not found" 
    raise ValueError(err)

def xsp2co1_fast_order(wx, via_word = True):
    assert isinstance(wx, Xsp2_Co1)
    if not via_word:
        return chk_qstate12(xsp2co1_order_elem(wx._data))
    m = MM0(wx)
    return xsp2co1_order_word(m._data, m.length)

def xsp2co1_fast_half_order(wx):
    assert isinstance(wx, Xsp2_Co1)
    buf = np.zeros(10, dtype = np.uint32)
    m = MM0(wx)
    res = chk_qstate12(xsp2co1_half_order_word(m._data, m.length, buf))
    o, l = divmod(res, 256)
    assert 0 <= l <= 10
    out = Xsp2_Co1()
    chk_qstate12(xsp2co1_mul_elem_word(out._data, buf, l))
    return o, out


@pytest.mark.xsp2co1
def test_elem_order(ntests = 50, verbose = 0):
    neutral = Xsp2_Co1()
    print("Test function computation of order in G_x0")
    for i, w in enumerate(make_testwords(monomial=False, ntests=ntests)):
        wx = Xsp2_Co1(w)
        o_ref = xsp2co1_ref_order(wx)
        o = xsp2co1_fast_order(wx, via_word = i & 1)
        ok = o == o_ref
        if verbose or not ok:
            print("Test", i+1)
            print("g = ", MM(*w))
            print("order =", o)
            if not ok:
                print("expected:", o_ref)
                raise ValueError("Computation or order has failed")

        o1, invol = xsp2co1_fast_half_order(wx)
        assert o1 == o
        if (o & 1):
            assert invol == neutral
        else:
            assert xsp2co1_fast_power(wx, o >> 1) == invol
            assert invol * invol == neutral

#######################################################################
# benchmark
#######################################################################

def one_benchmark(ntests):
    tags = "dxyplplplplplp"
    samples = [Xsp2_Co1(*rand_G_x0_testword(tags)) for i in range(400)]
    a = np.zeros(10, dtype = np.uint32)
    t_start = time.process_time()
    for elem in samples:
        elem_data = elem._data
        for i in range(ntests):
            assert xsp2co1_elem_to_word(elem_data, a) >= 0
    t = time.process_time() - t_start
    return t / (ntests * len(samples))


def benchmark():
    t = 1.0e6 * one_benchmark(100)
    print("\nTime for function xsp2co1_elem_to_word: %.3f us" % t)


#######################################################################
# Main program 
#######################################################################

if __name__ == "__main__":
    #test_monomial_to_word()
    #test_elem_to_word()    
    benchmark()
    
    


import os
import subprocess
from optparse import OptionParser

parser = OptionParser()


ROOT_DIR = os.path.dirname(os.path.realpath(__file__))


PROGRAM_FILES = [".c", ".h", ".pxd", ".pxi", ".pyx"]

if os.name == "posix":
   EXTENSIONS = [".so"]
elif os.name == "nt":
   EXTENSIONS = [".pyd"]
else:
   EXTENSIONS = []

EXTENSION_STR = ", ".join(EXTENSIONS) if len(EXTENSIONS) else "?"



def del_pyc(verbose = False):
    if verbose:
        print("\nDeleting python intermedate files (.pyc)") 
    for root, dirs, files in os.walk(ROOT_DIR, topdown=False):
        for name in files:
            if os.path.splitext(name)[1] == ".pyc":
                path = os.path.join(root, name)
                try:
                    if verbose:
                       print("Delete %s" % path)
                    os.remove(path)
                except:
                    if verbose:
                       print("failed")


def del_c(verbose = False):
    C_DIRS = [
        ["src", "mmgroup", "dev", "c_files"],
        ["src", "mmgroup", "dev", "pxd_files"],
    ]
    if verbose:
        print("\nDeleting automatically generated .c and .pxd files") 
    for dir_list in C_DIRS:
        try:
            dir = os.path.join(ROOT_DIR, *dir_list)
        except:
            continue
        files = os.listdir(dir)
        for name in files:
            if name == "readme.txt":
                continue
            path = os.path.join(dir, name)
            try:
                if verbose:
                   print("Delete %s" % path)
                os.remove(path)
            except:
                if verbose:
                   print("failed")


DATA_FILE_DICT = {
    ("src", "mmgroup", "dev", "mm_reduce" ): [
        "order_vector_data",
    ],
    ("src", "mmgroup", "tests", "test_axes"): [
        "sample_axes", "baby_sample_axes",
    ],
    ("src", "mmgroup", "tests", "test_involutions"): [
        "involution_samples",
    ]
}



def del_data(verbose = False):
    if verbose:
        print("\nDeleting automatically generated python data files") 
    for path_info, file_list in DATA_FILE_DICT.items():
        dir = os.path.join(ROOT_DIR, *path_info)
        for filename in file_list:
            path = os.path.join(dir, filename + ".py")
            try:
                if verbose:
                   print("Delete %s" % path)
                os.remove(path)
            except:
                if verbose:
                   print("failed")

def del_ext(verbose = False):
    if verbose:
        print("\nDeleting automatically generated python extensions") 
    if os.name == "nt":
        dir = os.path.join(ROOT_DIR, "src", "mmgroup")
        files = os.listdir(dir)
        for name in files:
            if os.path.splitext(name)[1] == ".pyd":
                path = os.path.join(dir, name)
                try:
                    if verbose:
                       print("Delete %s" % path)
                    os.remove(path)
                except:
                    if verbose:
                        print("failed")
    else:
        s = "Dont't know how do delete python extensions in '%s' system"
        print(s % os.name)
      

def git_checkout_data(verbose = False):
    if verbose:
        print("\nCheck out automatically generated data files with git") 
    for path_info, file_list in DATA_FILE_DICT.items():
        dir = os.path.join(*path_info)
        for filename in file_list:
            path = os.path.join(dir, filename + ".py")
            try:
                if verbose:
                   print("Checking out %s" % path)
                args = ["git", "checkout", path]
                print(" ".join(args))
                subprocess.check_call(args)
            except:
                if verbose:
                   print("failed")




def main():
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("-p",  dest="del_pyc", action="store_true",
        help="Delete intermediate python files (.pyc)")
    parser.add_option("-c",  dest="del_c", action="store_true",
        help="Delete automatically generated program files (%s)" %
        ", ".join(PROGRAM_FILES))
    parser.add_option("-d",  dest="del_data", action="store_true",
        help="Delete automatically generated data files (.py)")
    parser.add_option("-x",  dest="del_ext", action="store_true",
        help="Delete automatically generated extensions (%s)" %
        EXTENSION_STR)
    parser.add_option("-a",  dest="del_all", action="store_true",
        help="Delete all automatically generated files" )
    parser.add_option("-g",  dest="git_checkout", action="store_true",
        help="Checkout all automatically generated files with git" )
    parser.add_option("-v",  dest="verbose", action="store_true",
        help="Verbose operation" )
    
    options, args = parser.parse_args()
    verbose = options.verbose

    if options.del_pyc or options.del_all:
        del_pyc(verbose)
    if options.del_c or options.del_all:
        del_c(verbose)
    if options.del_data or options.del_all:
        del_data(verbose)
    if options.del_ext or options.del_all:
        del_ext(verbose)
    if options.git_checkout:
        git_checkout_data(verbose)


if __name__ == "__main__":
    main()





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix
from mmgroup.structures.qs_matrix import binary
from mmgroup.structures.qs_matrix import qs_pauli_matrix
from mmgroup.structures.qs_matrix import qs_unit_matrix
from mmgroup.structures.qs_matrix import pauli_vector_mul
from mmgroup.structures.qs_matrix import pauli_vector_exp



from mmgroup.clifford12 import bitmatrix64_t, bitmatrix64_mul
from mmgroup.clifford12 import bitmatrix64_rot_bits
from mmgroup.clifford12 import bitmatrix64_xch_bits
from mmgroup.clifford12 import bitmatrix64_echelon_l
from mmgroup.clifford12 import bitmatrix64_reverse_bits

#####################################################################
# Create test matrices
#####################################################################


def rand_unitary_matrix(n, scalar = True):
    m = qs_unit_matrix(n)
    e = randint(-16,16) if scalar else 0
    m.mul_scalar(e, randint(0,7))
    if n == 0:
        return m
    mask = (1 << n) - 1
    for i in range(n+2):
        v = randint(0, mask-1)
        vc =  randint(0, mask) & ~v
        m.gate_ctrl_not(vc, v)
        v1 = randint(0, mask)
        v2 =  randint(0, mask)
        m.gate_ctrl_phi(v1, v2)
    m.gate_phi(randint(0, mask), 1)
    m.gate_phi(randint(0, mask), 2)
    m.gate_h(randint(0, mask))
    m.gate_not(randint(0, mask))
    for i in range(n+2):
        v = randint(0, mask-1)
        vc =  randint(0, mask) & ~v
        m.gate_ctrl_not(vc, v)
        v1 = randint(0, mask)
        v2 =  randint(0, mask)
        m.gate_ctrl_phi(v1, v2)
    return m.echelon()

def rand_pauli_vectors(n, length):
    mask = (1 << (2*n+2)) - 1
    return [randint(0, mask) for i in range(length)]

def create_conjugate_data():
    yield rand_unitary_matrix(0), [0,1,2,3]
    yield qs_unit_matrix(1), [0,1,2,3]
    
    for n in range(1, 13):
        for i in range(10):
            m = rand_unitary_matrix(n)
            v = rand_pauli_vectors(n, max(10, n))
            yield m, v



#####################################################################
# Test method pauli_conjugate()
#####################################################################


    
@pytest.mark.qstate
def test_pauli_conjugate(verbose = 0):
    """Test the conjugation of Pauli matrix with unitary matrix"""
    for ntest, (m, v) in enumerate(create_conjugate_data()):
        n = m.shape[0]
        p = [qs_pauli_matrix(n, x) for x in v]
        pv = [x.pauli_vector() for x in p]
        ok = pv == v
        if verbose or not ok:
            mm = m.copy()
            print("\nTest %d: v = %s" % (ntest, binary(v[0],0,2*n+2)))
            print("m =", mm.reduce_matrix(), mm)
            print("Pauli matrix of v", p[0])
            if not ok:
                print("Input and recomputed Pauli vectors:")
                for i, x in enumerate(v):
                    bv =  binary(x,0,2*n+2)
                    bpv =  binary(pv[i],0,2*n+2)
                    print(bv, ",", bpv)
                if not ok:
                    err = "Error in recomputation of Pauli vector"
                    raise ValueError(err)
        mi = m.inv()
        assert m @ mi == qs_unit_matrix(n), (str(m), str(mi), str(m@mi))

        w = m.pauli_conjugate(v)
        w_ref = [ (m @ x @ mi).pauli_vector() for x in p]
        if verbose or w != w_ref:
            for i, w_i in enumerate(w):
                w_i_ref = w_ref[i]
                if w_i_ref != w_i:
                   s = "Pauli vector: %s, conjugated: %s, obtained: %s"
                   print (s % (hex(v[i]), hex(w_i_ref), hex(w_i)))
                   err = "Pauli vector conjugation error"
                   raise ValueError(err)
                elif verbose:
                   s = "Pauli vector: %s, conjugated: %s"
                   print (s % (hex(v[i]), hex(w_i)))

        w_noarg = m.pauli_conjugate(v, arg = False)
        if verbose:
            print("Without arg", [hex(x) for x in w_noarg])
        mask = (1 << (2 * n)) - 1
        assert w_noarg == [x & mask for x in w]



#####################################################################
# Test method to_symplectic()
#####################################################################

 

def bit_rev(x, k):
    x = int(x)
    y = sum( ((x >> (k-i-1)) & 1) << i  for  i in range(k) )
    x &= -(1 << k)
    return x + y
 



def py_to_symplectic(m):
    ERR_INV = "Qstate matrix is not imvertible"
    m.reduce()
    shape = m.shape 
    if shape[0] != shape[1]:
        raise ValueError(ERR_INV)
    n = shape[0]
    data = np.array(m.data, dtype = np.uint64)
    if len(data) <= n:
         raise ValueError(ERR_INV)
    if n == 0:
         return np.zeros((0,0),  dtype = np.uint64)
    data = data[1:]
    d_rows = len(data)
    AT = bitmatrix64_t(data, n)
    mask = 1 << (2*n - 1)
    sum_ = reduce(__or__, [int(data[i]) ^ (mask >> i) for i in range(n)], 0)
    if (sum_ >> n) & ((1 << n) - 1):
        raise ValueError(ERR_INV)
    bitmatrix64_xch_bits(data, 2 * n + 1, (1 << n) - 1)
    # Column order of ``data``
    #   Q_ker   |  A_cols   | |  A_rows  |  Q_rows  
    #  d_rows-n |     n     |1|    n     |     n
    r = bitmatrix64_echelon_l(data, d_rows, 2*n+1, d_rows)
    if r  != d_rows:
        raise ValueError(ERR_INV)
    S = AT >> n
    S = bitmatrix64_mul(S, data[n:])
    mask = (1 << (2*n)) - 1
    result = np.concatenate((S, data[:n])) & mask
    mask =  (1 << n) - 1
    for i in range(n):
        result[i] = int(result[i]) ^ (int(AT[i]) & mask)
    bitmatrix64_reverse_bits(result, n, 0)
    return result

 

@pytest.mark.qstate
def test_to_symplectic(verbose = 0):
    """Test mapping of unitary matrix to symplectic bit matrix"""
    for ntest, (m, v_) in enumerate(create_conjugate_data()):
        if verbose:
            print("\nTest", ntest+1, ", m =", str(m))
        n = m.shape[0]
        v = [1 << i for i in range(2*n)]
        p = [qs_pauli_matrix(n, x) for x in v]
 
        mi = m.inv()

        w = [int(x) for x in py_to_symplectic(m)]

        ok = True
        mask = (1 << 2*n) - 1
        w_ref = [ (m @ x @ mi).pauli_vector() & mask  for x in p]
        if verbose or w != w_ref:
            for i, w_i in enumerate(w):
                w_i_ref = w_ref[i]
                if w_i_ref != w_i:
                   s = "Pauli vector: %s, conjugated: %s, obtained: %s"
                   print (s % (hex(v[i]), hex(w_i_ref), hex(w_i)))
                   ok = False
                elif verbose:
                   s = "Pauli vector: %s, conjugated: %s"
                   print (s % (hex(v[i]), hex(int(w_i))))
        if not ok:
            print("Qstate matrix", str(m))
            err = "Pauli vector conjugation error"
            raise ValueError(err)

        w_c = m.to_symplectic()
        #if len(w_c): w_c[0] += 1
        assert list(w_c) == w_ref, (w_c, w_ref)
      

#####################################################################
# Create test matrices for Pauli group operation
#####################################################################


def create_pauli_vectors():
    yield 0, 2, 3
    yield 4, 0x1, 0x10
    yield 16, 0x100000001, 0x200010000
    for n in range(0,21):
        for i in range(4):
            v = rand_pauli_vectors(n, 2)
            yield n, v[0], v[1]

@pytest.mark.qstate
def test_pauli_multiplication(verbose = 0):
    for ntest, (n, v1, v2) in enumerate(create_pauli_vectors()):
        p1 = qs_pauli_matrix(n, v1)
        p2 = qs_pauli_matrix(n, v2)
        # Check product v1 * v2
        v3 = pauli_vector_mul(n, v1, v2)
        p3 = qs_pauli_matrix(n, v3)
        v3_ref = (p1 @ p2).pauli_vector()
        ok = v3 == v3_ref
        if verbose or not ok:
            print("Test %d, dim = %d" % (ntest, n))
            print("v1 = %s, v2 = %s, v1 * v2 = %s" % 
                (hex(v1), hex(v2), hex(v3_ref)) )
            if not ok:
                print("m1", p1)
                print("m2", p2)
                print("m1 * m2", p1 @ p2)
                print("pauli vector obtained: ", hex(v3))
                err = "Pauli vector multiplication failed"
                raise ValueError(err)
        # Check powers of v1
        p_e_list = [qs_unit_matrix(n), p1, p1 @ p1, p1.H]
        v_e_ref_list = [m.pauli_vector() for m in p_e_list]
        v_e_list = [pauli_vector_exp(n, v1, e) for e in range(4)]
        ok = v_e_list == v_e_ref_list
        if verbose or not ok:
            print("v1 = ", hex(v1))
            for e, v_ref in enumerate(v_e_ref_list):
                if ok:
                    print("v1 ** %d = %s" % (e, hex(v_ref)))
                else:
                    print("v1 ** %d = %s, obtained: %s, v1 = %s" % 
                         (e, hex(v_ref), hex(v_e_list[e]), v1))
            if not ok:
                err = "Pauli vector exponentiation failed"
                raise ValueError(err)
        



#####################################################################
# Test matrix exponentiation and order computation
#####################################################################


def create_exp_data():
    yield qs_unit_matrix(2), 0
    yield qs_unit_matrix(4), -3
    for n in range(1, 13):
        for i in range(10):
            m = rand_unitary_matrix(n)
            m.mul_scalar(randint(-2, 2))
            e = randint(-10000, 10000)
            yield m, e
   


def ref_power(m, e):
    if e > 1:
        m1 = ref_power(m, e >> 1)
        m1 = m1 @ m1
        return m1 @ m  if e & 1 else m1 
    if e == 1:
        return m
    if e == 0:
        return qs_unit_matrix(m.shape[0])
    if e < 0:
        mi = m.inv()
        assert m @ mi == qs_unit_matrix(m.shape[0])
        return  ref_power(mi, -e)
    

def ref_trace(m):
    mc = m.complex()
    norm = np.amax(abs(mc))
    tr = np.trace(mc)
    if norm > 0 and abs(tr / norm) < 1.0e-8:
        return 0
    return np.trace(mc)

@pytest.mark.qstate
def test_matrix_power(verbose = 0):
    MAX_ORDER = (2**8-1)*(2**6-1)*2**10
    """Test matrix exponentiation and commputation of trace"""
    for ntest, (m, e) in enumerate(create_exp_data()):
        # Test exponentiation
        me = m.power(e)
        me_ref = ref_power(m, e)
        ok = me == me_ref
        if verbose or not ok:
            mm = m.copy()
            print("\nTest %d: exponent = %d, " % (ntest, e))
            print("m =", m.reduce())
            print("Result of exponentiation:", me_ref)
            if not ok:
                print("Obtained:", me)
                raise ValueError("Matrix exponentiation failed")
        nqb = m.shape[0]
        m.mul_scalar(-m.lb_norm2())
        assert m.H == m.inv()

        # Test computation of trace
        if nqb > 8:
            continue
        tr = m.trace()
        tr_ref = ref_trace(m)
        if tr == 0:
            ok = tr_ref == 0
        else:
            ok = abs(tr_ref / tr - 1) < 1.0e-6
        if verbose or not ok:
            print("\nTrace of m:",  tr)
            if not ok:
                print("m =", m.reduce())
                if nqb < 3:
                    print("m =\n", m.complex())
                print("\nTrace expected:",  tr_ref)
                s, f = m._trace_factor()
                print("Low level trace result:", s, hex(f))
                print("Intermediate")
                m1 = m.copy()
                for i in range(nqb):
                     m1.gate_ctrl_not(1 << i, 1 << (nqb + i));
                print(m1)
                m1 = m1.restrict(nqb, nqb)
                print(m1)
                raise ValueError("Computation of trace failed")

        # Testing order 
        if nqb > 4:
            continue
        order = m.order(MAX_ORDER)
        if verbose:
            s = print("Scaled matrix m has order %d" % order)
        assert m.power(order) == qs_unit_matrix(m.shape[0])

                



r"""This is yet a stub


"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from mmgroup.bitfunctions import bw24, lmap, pivot_binary_high
from operator import __or__, __xor__

import random
import numpy
from functools import reduce
from numpy import array, zeros, uint8, uint16, uint32

from mmgroup.dev.mat24.mat24_ref import Mat24




#######################################################################
# Auxiliary functions 
#######################################################################



def list_to_vect(list_):
    return  reduce(__or__, [1 << x for x in list_], 0) 


def y_table(g, cocodes, sign, code_omega, cocode_x, verbose = 0):
    """Return table for function  ``gen_leech2_reduce_n`` 


    This table is used to map an entry :math:`x_d x_\delta`,
    where :math:`d` is a (signed) octad or dodecad, and 
    where :math:`\delta` is even, to a standard form. The
    table is to be used for a fixed Golay code element 
    :math:`d` up to sign and up to a factor :math:`\Omega`.

    Such a mapping is done via conjugation with an element
    of shape  :math:`y_e x_\epsilon`, with math:`e` a Goly
    code word math:`\epsilon` an even cocode word..

    Here :math:`d` is given by parameter ``g``. Thee we choose
    a Golay code word ``e`` with :math:`g \cap e = \delta`. 
    Using the returned table, ``e`` may be calculated as

    :math:`e = \sum_{i=0}^10 \delta_i \cdot t_i` ,

    where :math:`\delta_i` is the ``i``-th bit of
    :math:`\delta`, and :math:`t_i` is the entry ``table[i+1]`` 
    of the table.
    
    ``table[0]`` is equal to :math:`\theta(d)`. ``table[12]`` 
    is a sign bit ``s``. If the bit 23 of :math:`v \cdot y_e`
    differs from ``s`` then we have to replace :math:`e` by
    :math:`e + e'`, where :math:`e'` is given by ``table[13]``. 
    This operation adds  :math:`x_\Omega` to  :math:`x_d` in
    the cases relevant for function ``gen_leech2_reduce_n``.
    If :math:`v \cdot y_{e + e'}` has a sign bit set then
    we have to multiply that vector with  :math:`x_\epsilon`,
    where the cocode word :math:`epsilon` is given by 
    ``table[13]``. Entries ``table[11,12,13]`` are copied
    from parameters ``sign, code_omega, cocode_x``,
    respectively.

    Parameter ``cocode_x`` is an list of cocode words
    that will be meapped to zero by applying the table.

    On input, all Golay code and cocode word must be given
    as list of integers representing the bit positions set.

    In the returned table, all Golay code or cocode words
    are given as bit vector in **Golay code** or **cocode**
    representation, respectively.
    """
    gv = Mat24.vect_to_gcode(list_to_vect(g))
    b = []
    for i in range(11):
        b.append((Mat24.ploop_cap(gv, 1 << i) << 12) + (1 << i))
    for c in cocodes:
        b.append(Mat24.vect_to_cocode(list_to_vect(c)) << 12)
    if verbose:
        print("Function y_table, g =", hex(list_to_vect(g)))
        print("cocodes =", cocodes)
        print("cocode_x =", cocode_x)
        print("b =", [hex(x) for x in b])
    b, columns = pivot_binary_high(b)
    if verbose:
        print("reduced b =", [hex(x) for x in b])
        print("columns =", columns)
    table = [Mat24.ploop_theta(gv)]
    table1 = []
    for i in range(10, -1, -1):
        if columns[0] == i + 12:
            table1.append(b[0] & 0xfff)
            b, columns = b[1:], columns[1:]
        else:
            table1.append(0)
    table1.reverse()
    table += table1
    table.append(sign)
    table.append(Mat24.vect_to_gcode(list_to_vect(code_omega)))
    table.append(Mat24.vect_to_cocode(list_to_vect(cocode_x)))
    if verbose:
        print("table =", [hex(x) for x in table])
    return table

#####################################################################
# Tables used in the function to be tested
#####################################################################

STD_OCTAD = list(range(8))
CPL_OCTAD = list(range(8,24))
OMEGA = list(range(24))
STD_DODECAD = [0,4,8, 13,14,15, 17,18,19, 21,22,23]
CPL_DODECAD = [1,2,3, 5,6,7, 9,10,11, 12,16,20]

TABLE_OCTAD = y_table(STD_OCTAD, 
    [[0, 8], [8, 9]], 
    1,
    range(8, 24, 2),
    [0,12]
)
TABLE_DODECAD = y_table(STD_DODECAD, 
    [[0,12]],
    1,
    CPL_DODECAD,
    [0,16]
)


#####################################################################
# Expected results of function gen_leech2_reduce_n
#####################################################################


EXPECTED = {
      0:  ([],[]),
   0x20:  ([], [2,3]),
   0x21:  ([], [0]),
   0x22:  (STD_OCTAD, []),
   0x31:  (OMEGA, [0]),
   0x34:  (STD_OCTAD, [0,8]),
   0x33:  (OMEGA, [1,2,3]),
   0x36:  (STD_DODECAD, []),
   0x40:  ([], [0,1,2,3]),
   0x42:  (CPL_OCTAD, []),
   0x43:  ([], [1,2,3]),
   0x44:  (STD_OCTAD, [8,9]),
   0x46:  (STD_DODECAD, [0,12]),
   0x48:  (OMEGA, []),
}

def vector(gcode, cocode):
    gc = Mat24.vect_to_gcode(sum(1 << x for x in gcode))
    coc = Mat24.vect_to_cocode(sum(1 << x for x in cocode))
    return (gc << 12) ^ coc ^ Mat24.ploop_theta(gc)




MAP_VECTOR = dict([(s, vector(*y)) for s, y in EXPECTED.items()])
MAP_VECTOR[0] = 0x1000000


#######################################################################
# GenLeechReduceY
#######################################################################



class GenLeechReduceY(object):
    """This is yet a stub

    """

    tables = {
        "LEECH_RED_STD_OCTAD": STD_OCTAD,
        "LEECH_RED_CPL_OCTAD": CPL_OCTAD,
        "LEECH_RED_TABLE_OCTAD": TABLE_OCTAD,

        "LEECH_RED_STD_DODECAD": STD_DODECAD,
        "LEECH_RED_CPL_DODECAD": CPL_DODECAD,
        "LEECH_RED_TABLE_DODECAD": TABLE_DODECAD,

        "LEECH_RED_TABLE_MAP_SUBOCTADS": MAP_VECTOR.keys(),
        "LEECH_RED_TABLE_MAP_VECTORS": MAP_VECTOR.values(),

    }

    directives = {}





r"""Module ``mat24_ref`` contains the class ``Mat24.`` 

Class ``Mat24`` in module ``mmgroup.dev.mat24.mat24_ref`` is the 
table-providing class used by the code generator to generate the C 
file ``mat24_functions.c``. That C file contains the functionality
of the python extension ``mmgroup.mat24``.
Class ``Mat24`` may also be used as a table-providing class for 
generating other C files. 

Class ``Mat24`` also exports the same functions as the ``mmgroup.mat24``
extension as class methods. So it can be used for testing that extension 
and also as a pure-python substitute for that extension. 

Class ``Mat24`` is based on class ``mat24tables.Mat24Tables``. It also
uses classes, functions and tables from the following modules 
in ``mmgroup.dev.mat24``:

 *  ``make_addition_table``, ``make_mul_transp``,
    ``mat24aux``, ``mat24heptad``, ``mat24theta``

"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals
from functools import reduce


from random import randint
import types
import sys
import re
import os
import subprocess
from operator import __or__, __xor__

import numpy


from mmgroup.bitfunctions import bw24, bit_mat_mul, bitparity
from mmgroup.bitfunctions import pivot_binary_low

from mmgroup.dev.mat24.mat24tables import Mat24Tables

from mmgroup.dev.mat24.mat24aux import Lsbit24Function, MatrixToPerm

from mmgroup.dev.mat24.mat24heptad import HeptadCompleter

from mmgroup.dev.mat24.mat24theta import make_augmented_theta_table
from mmgroup.dev.mat24.mat24theta import make_autpl_qf_table

from mmgroup.dev.mat24.make_addition_table import BitMatrixMulFix

from mmgroup.dev.mat24.make_mul_transp import BitMatrixMulTransp

from mmgroup.dev.mat24.mat24_doc import __doc__ as Mat24__doc__


from mmgroup.generate_c import UserDirective, UserFormat

from mmgroup.dev import config


TABLE_NAMES = None

###########################################################################
###########################################################################
# The Golay code (a pure python implementation)
###########################################################################
###########################################################################


class Mat24(Mat24Tables): 
    r"""Provide functions for the Mathieu group Mat24 and the Parker loop.

    These functions are exported as class methods. They are considered as
    pure python substitutes for the functions in the ``mmgroup.mat24``
    extension.

    So they can be used in an early stage of the build process where that
    extension is not yet available. The are also used for testing the
    ``mmgroup.mat24`` package.

    See module ``mmgroup.dev.mat24.mat24_doc`` for an overview of the
    functionality of the ``mmgroup.mat24`` extension. 

    This class is also a table-providing class used by the code
    generator for generating C code containing the functionality
    of the  ``mmgroup.mat24`` extension. For details, see section 
    *How the code generator is used* in 
    *The mmgroup guide for developpers*. 
    """
    ## Create tables for faster computetion
    MAT24_ORDER =  244823040 
    matrix_to_perm_ = MatrixToPerm(Mat24Tables.basis[12:])
    heptad_completer = HeptadCompleter(Mat24Tables)
    theta_table = make_augmented_theta_table()
    autpl_qf_table = make_autpl_qf_table(theta_table)
    autpl_qf_table64 = make_autpl_qf_table(theta_table, 64)
    verbose = False
    basis_weights_8 = Mat24Tables.basis_weights_8()

    # recip_basis(i & 31) shall not fail in C 
    recip_basis_c = numpy.append(Mat24Tables.recip_basis.copy(), [0]*8)

    ## Collect tables and coding functions for generating C code
    tables = {
           "Mat24_enc_table0"    : Mat24Tables.enc_table0,
           "Mat24_enc_table1"    : Mat24Tables.enc_table1,
           "Mat24_enc_table2"    : Mat24Tables.enc_table2,
           "Mat24_dec_table0"    : Mat24Tables.dec_table0,
           "Mat24_dec_table1"    : Mat24Tables.dec_table1,
           "Mat24_dec_table2"    : Mat24Tables.dec_table2,
           "Mat24_basis"         : Mat24Tables.basis,
           "Mat24_recip_basis"   : recip_basis_c,
           "Mat24_syndrome_table"    : Mat24Tables.syndrome_table,
           "Mat24_oct_enc_table"     : Mat24Tables.oct_enc_table,
           "Mat24_oct_enc_offset"    : Mat24Tables.oct_enc_offset,
           "Mat24_oct_dec_table"     : Mat24Tables.oct_dec_table, 
           "Mat24_theta_table"       : theta_table,
           "Mat24_autpl_qf_table"    : autpl_qf_table,
           "Mat24_autpl_qf_table64"  : autpl_qf_table64,
           "Mat24_doc"               : Mat24__doc__, 
           "INT_BITS"                : config.INT_BITS,
           "Mat24_basis_weights_8"   : basis_weights_8,
    }

    directives = {}

    _subgenerators = [
            Lsbit24Function, 
            matrix_to_perm_,
            heptad_completer,
            BitMatrixMulFix(),
            BitMatrixMulTransp(),
    ]

    for gen in _subgenerators:
        tables.update(gen.tables())
        directives.update(gen.directives())


    @classmethod
    def str_basis(cls, with_reciprocal_basis = False):
        def show(text, basis):
            s = text + "\n"
            for i,x in enumerate(basis): 
                if i % 6 == 0: s += "    "
                s +=  "%06x " % x
                if i % 6 == 5: s += "\n"
            return s
        s = "We list the basis vectors of the Golay code and of its cocode."
        s += """

Basis vectors have indices 0,...,11. Each basis vector is displayed 
as a hexadecimal number with bit i (of valence 2**i) corresponding  
to component i of the basis vector in GF(2)^24 for i = 0,...,23.
Golay cocode vectors are to be understood modulo the Golay code. 

"""
        s +=  show( "Golay cocode basis", cls.basis[:12] )
        s +=  show( "Golay code basis" , cls.basis[12:] )
        if with_reciprocal_basis:
             s += show( "Reciprocal basis", cls.recip_basis )
        return s
      
    @classmethod
    def show_basis(cls, with_reciprocal_basis = True):
        print(cls.str_basis(with_reciprocal_basis))

    ###########################################################################
    # Parker Loop
    ###########################################################################

    @classmethod
    def ploop_theta(cls, v1):
        """Return the theta function for the Parker loop.

        theta is a quadratic from from the Golay code C to the cocode C*.
        Here parameter v1 of function theta is represented as a Golay code 
        word. The result of the function is represented as a Golay cocode 
        word. The cocycle of the Parker loop is given by:

             cocycle(v1,v2) =   scalar(theta(v1), v2)
        
        with  scalar(.,.) the scalar product.
        """
        return cls.theta_table[v1 & 0x7ff] & 0xfff

    @classmethod
    def ploop_cocycle(cls, v1, v2):
        """Return the cocycle of the Parker loop.

        Then the Parker Loop product is given by

             v1 (*) v2  =  v1 ^ v2 * (-1)**cocycle(v1, v2) . 
        """
        s = cls.ploop_theta(v1) & v2 & 0xfff
        s ^= s >> 6
        s ^= s >> 3
        s = 0x96 >> (s & 7)
        return s & 1 


    @classmethod
    def mul_ploop(cls, v1, v2):
        """Return the Parker loop product v1 (*) v2

        Here v1 and v2 are integers coded as follows:
        bit 0,...,11:   representation as Golay code word
        bit 12:         Parker loop sign
        otther bits:    ignored
        """

        return v1 ^ v2 ^ (cls.ploop_cocycle(v1, v2) << 12)


    @classmethod
    def pow_ploop(cls, v1, u_exp):
        """Return power v1 ** u_exp of the Parker loop element v1."""
        return (v1 & -(u_exp & 1)) ^ (
             cls.theta_table[v1 & 0x7ff] & ((u_exp & 2) << 11) )


    @classmethod
    def ploop_comm(cls, v1, v2):
        """Return commutator of Golay code word v1 and v2.

        This is 0 if the intersection of the vectors v1 and v2 has
        bit weight 0 mod 4 and 1 is that intersection has bit weight 
        2 mod 4. v1 and v2 are in 'gvect' or 'ploop' representation.
        """
        cap = cls.gcode_to_vect(v1) & cls.gcode_to_vect(v2)
        return (cls.bw24(cap) >> 1) & 1

    @classmethod
    def ploop_assoc(cls, v1, v2, v3):
        """Return associator of Golay code words v1, v2 and v3
 
        This the parity of the intersection of the vectors v1, v2 and 
        v3.  v1, v2 and v3 are in 'gvect' or 'ploop' representation.
        """
        assoc = (cls.gcode_to_vect(v1) & cls.gcode_to_vect(v2)
                & cls.gcode_to_vect(v3))
        return cls.bw24(assoc) & 1


    @classmethod
    def ploop_cap(cls, v1, v2):
        """Return intersection of two Golay code words as cocode word.

        v1 and v2 are in 'gvect' or 'ploop' representation, the result
        is returned in 'cocode' representation.
        """
        return ( cls.theta_table[(v1 ^ v2) & 0x7ff] 
                  ^  cls.theta_table[v1 & 0x7ff] 
                  ^  cls.theta_table[v2 & 0x7ff]  ) & 0xfff

    @classmethod
    def ploop_solve(cls, a):
        """Return cocode element that kills signs of Parker loop elements
 
        Here 'a' is an array of Parker loop elements. The function tries 
        to find a cocode element that makes all these Parker loop 
        elements positive, when operating on them as a diagonal 
        automorphism. The function returns the least cocode element in 
        lexical order satisfying that condition. For that order we 
        assume that lower bits have higher valence.
        If no such cocode element exists, ValueError is raised.
        """
        a1 = [x & 0x1fff for x in a]
        basis, columns = pivot_binary_low(a1)
        res = 0
        for b, col in zip(basis, columns):
            res |= ((b >> 12) & 1) << col 

    ###########################################################################
    # Mathieu group M24: conversion between representations
    ###########################################################################
    


    @classmethod
    def perm_complete_heptad(cls, p_io):
        """Complete a permutation p given by p_io to an element of Mat24.

        p must be a list of length 24. Entries p[i], i = 0,1,2,3,4,5,8.
        must make up a valid umbral heptad, i.e. a heptad not contained 
        in an octad. p[0],...,p[5] must be contained in an octad, p[8]
        must not be contained in that octad. The other entries of 
        input p are ignored.

        It can be shown that such a permutation p can be completed to 
        a unique element of Mat24.

        The function returns 0 in case of success and raises ValueError
        otherwise. In case of success, p_io is completed to an element of
        the Mathieu group Mat24. 
        """
        if cls.heptad_completer.complete_heptad(p_io) != 0:
            raise ValueError("Could not complete permutation in Mat24")


    @classmethod
    def perm_check(cls, p1):
        """Check if permutation p1 is in in the Mathieu group Mat24.

        The function returns zero iff this is the case.
        """
        p2 = list(p1[:])
        if  cls.perm_complete_heptad(p2):
             return -1
        return  p2 != list(p1[:])

    @classmethod
    def perm_complete_octad(cls, p_io):
        """Complete an octad given by 6 elements of it. 
        
        This is a simplified version of function 
        perm_complete_heptad().

        We complete an octad from entries p_io[i], i = 0,1,2,3,4,5.
        We store the remaining entries of the octad in p_io[6],
        p_io[7]. The order of these two entries is such that there
        exists a permutation in Mat24 that maps from the (ordered) 
        standard octad  0,...,7  to p_io[0],...,p_io[7].
        If p_io[5] is None, that it is set to the lowest possible
        value such that there is an octad  p_io[0],...,p_io[7].
        """
        assert len(p_io) >= 8
        p = list(p_io[:8]) + [None] * 16
        v = sum([1 << i for i in p[:5]])
        syn = cls.syndrome(v)
        if p_io[5] in (24, None):
            p[5] = cls.lsbit24(syn)
        p[8] = cls.lsbit24((v | syn) ^ 0xffffff)
        cls.perm_complete_heptad(p)
        p_io[6:8] = p[6:8]

    @classmethod
    def perm_from_heptads(cls, h1, h2):
        """Try to find a permutation p that maps heptad h1 to h2

        h1 and h2 must be lists of length 7 defining two umbral heptads,
        i.e. heptads not contained in an octad. If a permutation p in
        the Mathieu group Mat24 that maps h1 to h2 exists, it is unique. 

        The function returns p if such a p exists an is unique and
        it returns None otherwise.
        """
        return cls.heptad_completer.perm_from_heptads(h1, h2)

    @classmethod
    def m24num_to_perm(cls, u_m24):
        """Return permutation with number u_m24 in the Mathieu group Mat24.

        The inverse of this function is member function Mat24_perm_to_int()
        This is just a short and convenient way to number elements of Mat24.
        Input u_m24 = 0 gives the identity permutation.

        0 <= u_m24 < 244823040 = order(Mat24) must hold.

        For internal operation see
        mat24Heptad.HeptadCompleter.int_to_perm
        """
        p = cls.heptad_completer.int_to_perm(u_m24)
        return p

    @classmethod
    def perm_to_m24num(cls, p1):
        """Convert permutation p1 in the Mathieu group Mat24 to an integer.

        This reverses member function int_to_perm(). The input permutation
        is not checked.
        """
        return cls.heptad_completer.perm_to_int(p1[:])
            

    @classmethod
    def perm_to_matrix(cls, p1):
        """Convert a permutation p1 in  Mat24 to a matrix.

        The matrix is a 12 x 12 bit matrix acting on the Golay code
        vectors by right multiplication.

        Permutation p is not checked to be a member of the Mathieu group.
        """
        mat = [cls.recip_basis[p1[i]] >> 12 for i in range(24)]
        return bit_mat_mul(cls.basis[12:], mat)


    @classmethod
    def matrix_to_perm(cls, m1):
        """Convert element of Mat24 from matrix to permutation.

        The matrix m1 is a 12 x 12 bit matrix acting on the Golay code
        vectors by right multiplication. The matrix is not checked.
    
        The permutation is represented as a list.
        """
        basis = [cls.gcode_to_vect(v & 0xfff) for v in m1]
        return cls.matrix_to_perm_.compute(basis)


    #####################################################################
    # Mathieu group M24: operation of group elements
    #####################################################################

    @classmethod
    def op_vect_perm(cls, v1, p1):
        """Apply a permutation p1 to a vector v1 in GF(2)**24

        Here p1 is the permutation that maps i to p1[i]  for i=0,...,23.
        """      
        return sum([ ((v1 >> i) & 1) << p1[i] for i in range(24) ])


    @classmethod
    def op_gcode_matrix(cls, v1, m1):
        """Apply the 12 x 12 bit matrix m1 to a Golay code vector

        Here application means right multiplication v1 * m1. The code
        vector v1 is given in 'gcode' representation.
        """    
        return reduce(
                __xor__, [x for i,x in enumerate(m1) if v1 & (1<<i)], 0 )


    @classmethod
    def op_gcode_perm(cls, v1, p1):
        """Apply a permutation p1 to a Golay code vector v1
  
        Here p1 is the permutation that maps i to p1[i] for i=0,...,23,
        representing an element of the Mathieu group M24.
        
        Golay code vector v1 is given in gcode representation.
        """    
        v1 = cls.gcode_to_vect(v1)
        v1 = cls.op_vect_perm(v1, p1)
        return  cls.vect_to_vintern(v1) >> 12


    @classmethod
    def op_cocode_perm(cls, c1, p1):
        """Apply a permutation p to a Golay cocode vector c1

        Here p1 is the permutation that maps i to p1[i]  for i=0,...,23,
        representing an element of the Mathieu group M24.

        Golay cocode vector c1 is given in cocode representation.
        """
        v = c1
        y = - (((v >> 11) + 1) & 1)          # y = 0 if v is odd else -1
        v ^= cls.recip_basis[0] & y         # make v odd
        res = cls.recip_basis[p1[0]] & y     # .. and adjust result
        syn = cls.syndrome_table[v & 0x7ff ]  # get syndrome
        res ^=  cls.recip_basis[p1[syn & 31]] # do 1st syndrome vector
        syn = (syn >> 5) & 0x3ff             # mask out 2nd, 3rd synd. vector
        syn &=  ((syn + 0x100) >> 10) - 1    # kill syn if v >= 24*32
        res ^=  cls.recip_basis[p1[syn & 31]] ^ cls.recip_basis[p1[syn >> 5]]
                                             #  do 1st and 2nd syndrome vector
        return res & 0xfff


    @classmethod
    def mul_perm(cls, p1, p2):
        """Return p1 * p2, with p1, p2 in the Mathieu group M24

        p1, p2 are represented as permutations
        """
        return [ p2[p1[i]] for i in range(24) ]
 
    @classmethod
    def inv_perm(cls, p1):
        """Return inverse of p1, with p1 in the Mathieu group M24

        p1, is represented as a permutations
        """
        l = [None] * 24
        for i, x in enumerate(p1): l[x] =  i
        return l

             
    ###########################################################################
    # Automorphisms of the Parker Loop
    ###########################################################################



    @classmethod
    def autpl_set_qform(cls, m_io):
        """Recompute quadratic form on Parker loop automorphism m_io

        This functions augments the Parker loop automorphism m_io by
        a quadratic form qf. The form qf simplifies the application
        of m_io to Parker loop elements and also the multiplication
        of  Parker loop automorphisms. The form qf is stored in bits
        13,...,24 of the entries of m_io. 

        At present the documatation of the mathematical background
        behind that form is out of the scope of the module.
        """
        for i in range(12):
            qq = cls.ploop_theta(1 << i) & ((1 << i) - 1)
            s = cls.ploop_theta(m_io[i])
            for j in range(i):
                t =  s & m_io[j] 
                t ^= t >> 6
                t ^= t >> 3
                qq ^= ((0x96 >> (t & 7)) & 1) << j
            m_io[i] = (m_io[i] & 0x1fff) ^ (qq << 13)
        return m_io

    @classmethod
    def perm_to_autpl(cls, c1, p1):
        """Combine Mat24 and cocode element to Parker loop automorphism

        Given an element p1 of the Mathieu group Mat24 (in permutation 
        representation) and a Golay cocode element c1 (in cocode 
        representation), the function returns a Parker loop automorphism 
        m as a 12 x (12+13) matrix, i.e. in 'autpl' representation.
        m contains the 12 images of the basis vectors of the Parker loop
        and a quadratic form for simplfying its operation on Pl.
        """
        m = cls.perm_to_matrix(p1)
        for i in range(12):
            m[i] ^= ((c1 >> i) & 1) << 12
        return  cls.autpl_set_qform(m)

    @classmethod
    def cocode_to_autpl(cls,  c1):
        """Convert cocode element c1 to Parker loop automorphism

        Same as perm_to_autpl(c1, p), with p the identity permutation
        """
        return [(1 << i) + (((c1 >> i) & 1) << 12) for i in range(12)]

    autpl_to_perm = matrix_to_perm 

    @classmethod
    def autpl_to_cocode(cls, m1):
        """Extract cocode vector c from Parker loop automorphism m1

        Then m1 = perm_to_autpl(c, p), where p is the permutation
        obtained by calling autpl_to_perm(m1).

        Note that m1 = cocode_to_autpl(c) *  perm_to_autpl(0, p).
        """
        return sum( ((m1[i] >> 12) & 1) << i for i in range(12) )
       

    @classmethod
    def op_ploop_autpl(cls, v1, m1):
        """Apply Parker loop automorphism m1 to Parker Loop element v1

        Here m1 is a Parker loop autmorphism (in autpl representation)
        and v1 is a cocde vector (in cocode representation).
        The function returns the resluting cocode vector  v1 * m1.
        """
        v = v1
        t = v & 0x1000
        for i in range(12):
            t ^= m1[i] & -((v >> i) & 1) 
        v = (t >> 13) & v 
        v ^= v >> 6
        v ^= v >> 3 
        v = (0x96 >> (v & 7)) & 1
        return (t & 0x1fff) ^ (v << 12)

    @classmethod
    def mul_autpl(cls, m1, m2): 
        """Return product m1 * m2 of Parker loop automorphisms m1, m2

        Here  m1, m2  and the result is in 'autpl' representation.
        """
        m3 = [cls.op_ploop_autpl(m1[i], m2) for i in range(12)]
        return cls.autpl_set_qform(m3)

    @classmethod
    def inv_autpl(cls, m1): 
        """Return inverse of Parker loop automorphism m1

        Here m1 and the inverse of it is in 'autpl' representation.
        """
        p = cls.matrix_to_perm(m1)
        p = cls.inv_perm(p)
        mi = cls.perm_to_matrix(p)
        m0 = cls.mul_autpl(mi, m1)
        for i in range(12):
            assert mi[i] & -0x1000 == 0, map(hex, mi)
            assert m0[i] & 0xfff == 1 << i
            mi[i] ^= m0[i] & 0x1000
        return cls.autpl_set_qform(mi)

    @classmethod
    def perm_to_iautpl(cls, c1, p1):
        """Return (inverse of p1, inverse of element of perm_to_autpl(c1, p1))
     
        This will be optimized in the fast version.  
        """
        return cls.inv_perm(p1), cls.inv_autpl(cls.perm_to_autpl(c1, p1)) 





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint, shuffle, sample

from mmgroup import mat24
from mmgroup import GCode, Cocode, PLoop, PLoopZ
from mmgroup import Octad, GcVector 
from mmgroup import AutPL
from mmgroup import Parity




#####################################################################
# Generate random elements of AutPL
#####################################################################



def rand_u7():
    """Generate a random umbral heptad"""
    o = Octad(sample(range(24), 5))
    return sample(o.bit_list[:], 6) + sample(((~o).bit_list)[:], 1)

def rand_autpl_u7():
    """Generate random Permutation in Mat24 from umbral heptads"""
    return AutPL(0, zip(rand_u7(), rand_u7()), Cocode("r"))



#####################################################################
# Test generation of permutations
#####################################################################



def perm_mapping(length):
    """Create a mapping defining a unique permutation p in Mat24.

    Return a triple (src, dest, perm_num) with src, dest being
    lists of the given 'length' satisfying:

        p[src[i]] = dest[i] for 0 <= i < length .    

    for a unique permutation p in Mat24. p has the internal number 
    perm_num. 7 <= length < 24 must hold. The function does not 
    use any internal numbering or random process of the mmgroup 
    package for creating the permutation p.
    """
    p  = rand_autpl_u7()  # p is a random permutation in AutPL
    perm = p.perm
    assert length >= 7
    while True:
        src = sample(range(24), length)
        if length >= 9 or len(Cocode(src)) >= 2:
            # The the permutation is determined by its effect on src
            return  src, [perm[i] for i in src], p.perm_num
        



def create_perm_testvectors():
    for i in range(40):
        for length in range(7, 25):
            yield perm_mapping(length)





@pytest.mark.ploop
def test_create_group_permutation():
    for i, (src, dest, num) in enumerate(create_perm_testvectors()):
        p1 = AutPL(0, zip(src, dest))
        p1.check()
        p1c = p1.copy().check()
        assert p1c == p1
        assert p1.perm_num == num
        assert p1 == AutPL(0, dict(zip(src, dest))).check()
        assert p1 == AutPL(0, num).check()
        assert p1 == AutPL(p1).check()
        assert p1.cocode == 0
        perm = p1.perm
        assert isinstance(perm, list)
        assert min(perm) == 0 and max(perm) == 23
        for x, y in zip(src, dest):
            assert perm[x] == y
        assert p1 == AutPL(0, perm)
        coc = Cocode(randint(1,0xfff))
        p2 = AutPL(coc, 0) * AutPL(0, perm)
        p2.check()
        assert p2 == (AutPL(coc) * AutPL(0, perm)).check()
        assert p2 == AutPL(coc, perm).check()
        assert p2 != p1
        assert p2.perm_num == p1.perm_num
        assert p2.cocode == coc.cocode
        p2c = AutPL(0, perm) * AutPL(coc, 0).check()
        assert p2.perm_num == p1.perm_num
        assert Cocode(p2c.cocode) * p2 == coc
        assert p2 == AutPL(Cocode(p2.cocode), 0) * AutPL(0, p2.perm)


#####################################################################
# Test Group multiplication and inverse
#####################################################################

@pytest.mark.ploop
def test_group_op(n_cases = 100):
    print("")
    for i in range(n_cases):
        p1 = AutPL('r', 'r')
        p2 = rand_autpl_u7() 
        p2 *= AutPL(Cocode(randint(0, 0xfff)))
        p12 = (p1 * p2).check()
        assert p12.rep == mat24.mul_autpl(p1.rep, p2.rep)
        if i < 1:
            print(p1, "*", p2, "=", p12)
        p3 =  AutPL('r', 'r')
        assert (p1 * p2) * p3 == p1 * (p2 * p3)
        p1i = (p1 ** -1).check()
        assert p1 * p1i == p1i * p1 == AutPL()
        

#####################################################################
# Test Group operation on Parker loop
#####################################################################


def autpl_testwords(n_cases = 50):
    for i in range(n_cases):
        if i % 4 == 0:
            yield rand_autpl_u7()
        else:
            yield AutPL('r', 'r') 


def ploop_testvectors(n_cases = 50):
    for i in range(n_cases):
        yield PLoop(randint(0, 0x1fff))



@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_group_ploop(n_cases):
    print("")
    for i, (p, g) in enumerate(zip(ploop_testvectors(n_cases),
            autpl_testwords(n_cases))):
        pg = p * g
        if i < 1:
            print(p, "*", g, "=", pg)
        pg_ref = PLoop(mat24.op_ploop_autpl(p.ord, g.rep)) 
        assert pg == pg_ref
        assert p/4 == pg/4 == parity(len(p)/4)


#####################################################################
# Test Group operation on Cocode and on vector space
#####################################################################

def cocode_testvectors(n_cases = 50):
    for i in range(n_cases):
        yield Cocode(randint(0, 0xfff))

def get_testvectors(n_cases = 50):
    for i in range(n_cases):
        yield GcVector(randint(0, 0xffffff))


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_group_ploop(n_cases):
    print("")
    for i, (c, v, g) in enumerate(zip(cocode_testvectors(n_cases),
           get_testvectors(n_cases), autpl_testwords(n_cases))):
        cg = c * g
        vg = v * g
        if i < 1:
            print(c, "*", g, "=", cg)
            print(v, "*", g, "=", vg)
        cg_ref = Cocode(mat24.op_cocode_perm(c.ord, g.perm)) 
        assert cg == cg_ref
        assert len(cg) == len(c)
        vg_ref = GcVector(mat24.op_vect_perm(v.ord, g.perm)) 
        assert vg == vg_ref
        assert len(vg) == len(v)
        if len(v) & 1 == 0:
            assert v/2 == vg/2 == Parity(len(v) >> 1)
        if len(v) & 3 == 0:
            assert v/4 == vg/4 == Parity(len(v) >> 2)
        assert v % 2 == vg % 2 == Parity(len(v))




#####################################################################
# Test construction of group words with permutations and cocode
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_group_from_perm(n_cases):
    for i in range(n_cases):
        h1 = rand_u7()
        h2 = rand_u7()
        autp = AutPL(0, zip(h1, h2))
        assert autp == AutPL(0, dict(zip(h1, h2)))
        assert autp.perm == mat24.perm_from_heptads(h1, h2)
        assert autp.cocode == 0
        perm_num = autp.perm_num
        assert perm_num == mat24.perm_to_m24num(autp.perm)
        assert autp == AutPL(0, mat24.perm_from_heptads(h1, h2))
        assert autp == AutPL(autp)
        assert autp == AutPL(0, autp.perm_num)
        assert autp == AutPL(0, zip(h1, h2))
        coc_num = randint(1, 0xfff)
        coc = Cocode(coc_num)
        assert coc != Cocode(0)
        assert coc.cocode == coc_num
        im_coc = coc * autp
        assert type(im_coc) == type(coc)
        assert AutPL(im_coc) ==  AutPL(coc)**autp
        aut_cp = AutPL(coc) * autp
        assert aut_cp == AutPL(coc_num, perm_num) 
        if coc_num and perm_num:
            assert aut_cp.as_tuples() == [('d', coc_num), ('p', perm_num)]        
        assert autp * AutPL(im_coc) ==  aut_cp
        assert type(aut_cp) == type(autp)
        assert aut_cp.perm_num == perm_num 
        assert aut_cp.cocode == coc_num 
        assert Parity(aut_cp) == Parity(coc) 
        assert aut_cp.parity == coc.parity  == Parity(coc).ord
        assert autp == AutPL() * autp == autp *  AutPL()
        with pytest.raises(TypeError):
            autp * coc
        with pytest.raises(TypeError):
            autp * Parity(randint(0,9))
        with pytest.raises(TypeError):
            autp * randint(2,9)
        with pytest.raises(TypeError):
            randint(2,9) * autp
        with pytest.raises(TypeError):
            autp * PLoop(randint(0, 0x1fff)) 



#####################################################################
# Test generation of random elements
#####################################################################

        
@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [50])
def test_rand_group_word(n_cases):
    for i in range(n_cases):
        aut_cp = AutPL('r', 'r') 
        perm_num = aut_cp.perm_num
        coc_num = aut_cp.cocode 
        assert aut_cp == AutPL(coc_num, perm_num) 
        assert 0 <= perm_num < mat24.MAT24_ORDER
        assert 0 <= coc_num < 0x1000
        aut_cpe = AutPL('e', 'r') 
        assert aut_cpe.parity == 0
        aut_cpe = AutPL('o', 'r')
        assert aut_cpe.parity == 1



        


r"""We describe the representations of a vector in :math:`\rho_p`

The most important representation of a vector in :math:`\rho_p` is 
the *internal* representation. All operations of the monster group 
are performed on vectors in in :math:`\rho_p` in internal 
representation.

In the *internal* representation a vector is stored as an array of 
``247488`` entries, where each entry is a bit field representing 
an integer modulo ``p``. Some components of a vector are stored
twice in that array and some entries of the array are unused. This
special structure facilitates the implementation of the operations
of the monster group.

The entries of a vector are organized as tuples ``(tag, i0, i1)``. 
Here indices ``i0, i1`` refer  to a two-dimensional array as 
indicated in column ``Size`` of the  following table. As usual in 
C, entries with adjacent last index  ``i1`` are stored in adjacent 
locations. For a mathematical description of an entry 
``(tag, i0, i1)``, see section 
*The Representation of the Monster Group* in the API reference.


  .. table:: Array sizes for tags in the internal representation
    :widths: 10 24 24 24 18

    ===== ============= ============== ============= ===========
    Tag   Size          Space used     Remarks       Offset
    ===== ============= ============== ============= ===========
    ``A``   ``24 x 24``   ``24 x 32``  (1), (2)           ``0``
    ``B``   ``24 x 24``   ``24 x 32``  (1), (2), (3)    ``768``
    ``C``   ``24 x 24``   ``24 x 32``  (1), (2), (3)   ``1536``
    ``T``  ``759 x 64``  ``759 x 64``  (4),            ``2304``
    ``X`` ``2048 x 24`` ``2048 x 32``  (1),           ``50880``
    ``Z`` ``2048 x 24`` ``2048 x 32``  (1),          ``116416``
    ``Y`` ``2048 x 24`` ``2048 x 32``  (1),          ``181952``
    ===== ============= ============== ============= ===========

Remarks
  
  1. As indicated in column ``Space used``, an array of size 
     ``n`` times ``24`` is stored in a space reserved for
     an array of size ``n`` times ``32``. So the entry with
     index ``(tag, i0, i1)`` is stored at location
     ``Offset(tag) + 32 * i0 + i1``. Unused entries must
     be equal to zero.
    
  2. The entry given by ``(tag, i0, i1)`` must be equal to the 
     entry given by ``(tag, i1, i0)``. 
    
  3. The diagonal entry given by ``(tag, i0, i0)`` must be equal to 
     zero. 
   
  4. The entry with  index ``(T, i0, i1)`` is stored at location
     ``Offset(T) + 64 * i0 + i1``.

The entries of a vector in internal representation are stored
a one-dimensional array of integers of type ``uint_mmv_t``. Here
type ``uint_mmv_t`` may be one of the C integer types
``uint64_t`` or ``uint32_t``, depending on the value ``INT_BITS``
in the configuration file ``config.py``. There ``INT_BITS`` must
be ``64`` or ``32``. Several adjacent entries of a vector are
stored as bit fields in a single integer of type ``uint_mmv_t``.
Entries with a lower index are stored at bits with lower valence.


The number of bits in a bit field is always a power of two. So e.g. 
for ``p = 3`` we use ``2`` bits; for ``p = 7`` we use ``4`` bits 
with the highest bit unused. In case ``p = 2**k - 1``, legal values 
for an entry are  ``0,...,2**k - 1``, with ``2**k - 1`` equal to 
``0``. Thus negation of a value can be done by complementing all
``k`` bits of that value. Apart from negation, the matrices
corresponding to operations of the monster may add, subtract and 
half the entries of a vector (modulo ``p``). These operations can 
easily be done on several entries simultaneously by manipulating a 
just single integer of type ``uint_mmv_t``.


As indicated above, we reserve ``32`` entries for arrays of
integers modulo ``p``  with ``24`` entries. So we require about 
``25.7%`` more memory than necessary. In some cases we need this 
extra memory anyway. E.g. for ``p = 3`` a ``64``-bit integer may 
store ``32`` entries, so that there will always be a slack of 
``8`` entries when storing ``24`` entries. 

Function ``mm_aux_mmv_size(p)`` returns the number of integers 
of type  ``uint_mmv_t`` required for storing a vector in external
representation. 

When writing or calculating entries with tags ``A, B, C`` then
the high-level function for manipulating vectors in internal 
representation make sure that e.g. entries with indices
``(A, i0, i1)`` and ``(A, i1, i0)`` are always set to the same 
value. These functions also make sure that unused bits in a bit 
field and unused bit fields are set to zero. Note that a bit field 
with value zero may contain the value ``p`` instead.

Direct access to an entry of a vector ``mv`` of integers modulo
``p`` in internal representation is difficult. Function 
``mm_aux_get_mmv(p, mv, i)`` returns the entry with index ``i`` of 
the  vector ``mv``. Function ``mm_aux_put_mmv(p, value, mv, i)`` 
sets the entry with index ``i`` to the given ``value``. Here valid 
indices are ``0 <= i < 196884``. The order of the indices ``i`` is 
the order given by the *external representation* of a vector
in :math:`\rho_p`.


The *external* representation of a vector in :math:`\rho_p`

There is also a so-called *external representation* of a vector in R_p. 
This is used to facilitate the access to vectors by external modules. 
Here the vector is represented as an array of 196884 integers of type
uint8_t. Basis vectors are ordered similar to the ordering for the
internal representation, but here the entries are in one-to-one
correspondence with the basis vectors. In the external representation 
there are no unused or duplicated entries.

More precisely, the order of the entries is:

  .. table:: Order of entries in external representation
    :widths: 25 25 25 25

    ===============  ===========  ==============  ===========
    Entries          Condition    No of entries   Offset
    ===============  ===========  ==============  ===========
    ``(A, i0, i1)``  ``i0 = i1``        ``24``         ``0``
    ``(A, i0, i1)``  ``i0 > i1``        ``276``       ``24``
    ``(B, i0, i1)``  ``i0 > i1``        ``276``      ``300``
    ``(C, i0, i1)``  ``i0 > i1``        ``276``      ``576``
    ``(T, i0, i1)``                  ``759*64``      ``852``
    ``(X, i0, i1)``                 ``2048*24``    ``49428``
    ``(Z, i0, i1)``                 ``2048*24``    ``98580``
    ``(Y, i0, i1)``                 ``2048*24``   ``147732``
    ===============  ===========  ==============  ===========

Indices (``tag, i,j``) for ``tag = A, B, C``, ``i > j`` are ordered 
as follows:

   *  ``(1,0),``
   *  ``(2,0), (2,1),``
   *  ``(3,0), (3,1), (3,2),``
   *  ``...``
   *  ``(i,0), (i,1), ..., (i,i-1),``
   *  ``...``
   *  ``(24,0), (24,1), ..., (24,23).``

Function ``mm_aux_bytes_to_mmv()`` converts a vector from external  
to internal representation, Function ``mm_aux_mmv_to_bytes()`` does 
the inverse conversion.



The *sparse* representation of a vector in :math:`\rho_p`

The Python interface to vectors in :math:`\rho_p` is optimized
for readability and not for speed. Here a typical task is to read
and modify single entries of a vector. In the internal 
representation the coordinates of a vector are indexed by tuples
containing a string. Transferring tuples or strings from Python
to C is awful. Here we need a representation of a vector in
:math:`\rho_p` where a single entry of a vector can be stored
in an integer variable.

A vector in :math:`\rho_p` can be stored in the *sparse* 
representation. 
Here a vector is stored as an array of 32-bit integers, where each 
entry stands for a multiple of a basis vector. A component of
a vector is stored in the bit fields of an integer as a tuple
``(tag, i0, i1, value)``. Here the tuple  ``(tag, i0, i1)`` is as
in the external representation, and ``value`` is the value of the
coordinate of the vector corresponding to ``(tag, i0, i1)``.
Entries with coordinate zero may be dropped. A 32-bit integer
encodes a tuple ``(tag, i0, i1, value)`` in bit fields as shown 
in the following table.


   .. table:: Bit fields in the sparse representation
     :widths: 20 80

     ======   ======================================================
     Bits     Meaning
     ======   ======================================================
     27..25   ``Tag``: 
              ``A = 1``, ``B = 2``, ``C = 3``, ``T = 4``,
              ``X = 5``, ``Z = 6``, ``Y = 7``
     24..15   Index ``i0``
     13.. 8   Index ``i1``
      7.. 0   The ``value`` of the coordinate of the basis vector;  
              if the modulus ``p`` is ``2**k - 1``  then only 
              the lowest ``k`` bits are evaluated.
     ======   ======================================================
            
In a C function the length of a sparse representation of a vector 
must be given as a parameter to the function. 
The order of the entries is irrelevant in the sparse 
representation. A sparse representation generated by a C function
contains at most one entry for each tuple ``(tag, i0, i1)``.
On input, entries with several equal tuples ``(tag, i0, i1)`` are 
accepted. Unless stated otherwise, the corresponding values of
such equal tuples are added.

In an entry with tag ``A``, ``B``, or ``C`` generated by this module
we always have ``i0 >= i1``. The ``value`` of an entry generated by 
this module is always less than the modulus ``p``.

When reading an entry, a ``value`` with ``0 <= value <= p`` is
accepted. Entries with tag ``A``, ``B``, or ``C`` and ``i < j`` are 
also accepted. Illegal tags or indices are usually ignored on input.

"""



import time
from random import randint, sample
import numpy as np
import pytest

from mmgroup.clifford12 import bitvector32_heapsort
from mmgroup.clifford12 import bitvector64_heapsort
from mmgroup.clifford12 import bitvector32_sort
from mmgroup.clifford12 import bitvector64_sort

dtypes = {
  32 : np.uint32, 64 : np.uint64
}



def one_test_sort(sorter, a, verbose = 0):
    if verbose: 
        print("a=")
        print(a)
    sorted_a = np.copy(a)
    sorter(sorted_a, len(a))
    if verbose: 
        print("sorted")
        print(sorted_a)
    sorted_ref = np.sort(np.copy(a))
    if verbose: 
        #print("sorted ref")
        #print(sorted_ref)
        pass
    assert (sorted_a == sorted_ref).all(), (sorted_a, sorted_ref)



def sort_testdata(dtype = 32):
    d_t = dtypes[dtype]
    for i in range(1,40):
        yield   np.array( 0xfffffffe * np.random.rand(i), dtype = d_t)
  
    testdata = [ (4, 10), (10, 100), (100,10), (256, 20), (1299, 774), 
            (10, 10**9),
          (1000000, 2**dtype-1), 
    ]
    for length, maxval in testdata:
        yield   np.array( maxval * np.random.rand(length), dtype = d_t )


HEAP_SORTERS =  {32 : bitvector32_heapsort, 64 : bitvector64_heapsort}
SORTERS =  {32 : bitvector32_sort, 64 : bitvector64_sort}

def do_test_heapsort(dtype = 32, verbose = 0):
    sorter = HEAP_SORTERS[dtype]
    print("testing %d-bit heapsort...       " % dtype, end = "")
    for a  in sort_testdata(dtype): 
         one_test_sort(sorter, a, verbose = verbose)
    print("passed")

def do_test_sort(dtype = 32, verbose = 0):
    sorter = SORTERS[dtype]
    print("testing %d-bit bitvector sort... " % dtype, end = "")
    for a  in sort_testdata(dtype): 
         one_test_sort(sorter, a, verbose = verbose)
    print("passed")



@pytest.mark.qstate
def test_sorters(verbose = 0):
    for dtype in [32, 64]:
        do_test_heapsort(dtype, verbose)
        do_test_sort(dtype, verbose)



def do_test_benchmark_sort(n, nsamples = 100, n_repeat = 100, alg = 1):
    from mmgroup.clifford12 import bitmatrix32_test_sort
    if alg in (1,2):
        print("Runtime %ssort in micoroseconds, %d 32-bit integers, %s samples"
        % ("heap" if alg == 2 else "", n, nsamples))
    a = np.random.randint(0, 0xffffffff, n * nsamples, dtype=np.uint32)
    a = a.reshape((nsamples, n))
    #print(a)
    timings = []  #time of experiment in microseconds
    t_empty = bitmatrix32_test_sort(0, a[0], n_repeat)
    for i in range(nsamples):
        t1 = bitmatrix32_test_sort(alg, a[i], n_repeat)
        timings.append(1.0e6 * (t1 - t_empty) /  n_repeat)
    #print(timings)
    t_min, t_max = min(timings), max(timings)
    t_ave = sum(timings) / nsamples
    t_sigma = (sum((t - t_ave)**2 for t in timings) / (nsamples - 1)) ** 0.5
    print("ave: %.3f +- %.3f, min: %.3f, max: %.3f" % (
        t_ave, t_sigma, t_min, t_max))


@pytest.mark.bench
@pytest.mark.qstate
@pytest.mark.slow
def test_benchmark_sort(verbose = 0):
    print("")
    do_test_benchmark_sort(100, 100, 100, 1)
    do_test_benchmark_sort(100, 100, 100, 2)
    do_test_benchmark_sort(1000, 100, 100, 1)
    do_test_benchmark_sort(1000, 100, 100, 2)
    do_test_benchmark_sort(10000, 100, 10, 1)
    do_test_benchmark_sort(10000, 100, 10, 2)
    do_test_benchmark_sort(100000, 100, 1, 1)
    do_test_benchmark_sort(100000, 100, 1, 2)
    do_test_benchmark_sort(1000000, 10, 1, 1)
    do_test_benchmark_sort(1000000, 10, 1, 2)







from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from mmgroup.bitfunctions import bw24, lmap

import random
import numpy
from numpy import array, zeros, uint8, uint16, uint32

import pytest



from mmgroup.dev.mat24.mat24_ref import Mat24
from mmgroup.dev.generators.gen_xi_ref import GenXi
from mmgroup import generators as gen  
from mmgroup.generators import gen_leech2_mul

        

#######################################################################
# Test functions
#######################################################################




#######################################################################
# test_gen_xi_leech
#######################################################################


def xi_samples(ntests):
    data = [1,0x1000, 2, 0x100, 0x800001 ]
    for x in data:
         yield x  
    for i in range(ntests):
        yield random.randint(0, 0x1ffffff)



TEST_PAR = [(GenXi, 200), (gen, 2000)]

@pytest.mark.gen_xi
@pytest.mark.parametrize("gen_xi_class, ntests", TEST_PAR )
def test_gen_xi_leech(gen_xi_class, ntests):
    xi = gen_xi_class.gen_xi_op_xi 
    mul =  gen_leech2_mul

    xl_old =  [0,0,0]
    for x in xi_samples(ntests):
        xl = [x, xi(x,1), xi(x,2)]
        assert xl[2]  == xi(xl[1], 1) 
        assert xl[1]  == xi(xl[2], 2) 
        assert xi(xl[2], 1) == xi(xl[1], 2) == x 
        for i in range(1,3):
           assert mul(xl[i], xl_old[i]) == xi(mul(x, xl_old[0]), i)
        xl_old = xl


#######################################################################
# test_gen_xi_short
#######################################################################


def is_short_vector_code(x):
    box, code = x >> 16, x & 0x7fff
    if box in [4,5]:
        return code & 0x1f < 24
    if box == 3:
        return code < 24576 # this is 384 * 64
    if box == 2:
        return code < 23040 # this is 360 * 64
    if box == 1:
        if code >=  2496: # this is 2 * 24 * 32 + 15 * 64 
             return False
        if code >= 1536:
             return True 
        j = code & 0x1f
        if j >= 24:
             return False
        i = (code >> 5) % 24
        return i != j
    return False

def normalize_short_vector_code(x):
    box, sign, code = x >> 16, x & 0x8000, x & 0x7fff
    if box == 1 and code < 1536 and (code & 31) < 24:
        b, i, j = code >= 768, (code >> 5) % 24, code & 31
        i, j = max(i,j), min(i,j)
        return 0x10000 + sign + 768 * b + 32* i + j 
    return x
    
 


def rand_short_vector(xmin = 0x10000, xmax = 0x60000):
    while 1:
        x = random.randint(xmin, xmax-1)
        if is_short_vector_code(x):
            return x
        


def short_samples(ntests):
    data = [ 0x10030,  0x10001, 0x35000, 0x3ddb0 ]
    for x in data:
        yield x
    for i in range(max(10, ntests // 50)):
        x = rand_short_vector(0x10000, 0x10000+1536)
        yield x
        yield x | 0x8000
    for i in range(ntests):
        yield rand_short_vector()
 

TEST_PAR = [(GenXi, 200), (gen, 2000)]


          
@pytest.mark.gen_xi
@pytest.mark.parametrize("gen_xi_class, ntests", TEST_PAR )
def test_gen_xi_short(gen_xi_class, ntests):
    xi = gen_xi_class.gen_xi_op_xi 
    xi_short = gen_xi_class.gen_xi_op_xi_short 
    to_leech = gen_xi_class.gen_xi_short_to_leech
    to_short = gen_xi_class.gen_xi_leech_to_short
    for x in short_samples(ntests):
        xl = to_leech(x)
        x_norm = normalize_short_vector_code(x)
        assert to_short(xl) == x_norm, lmap(hex, [x, xl, x_norm, to_short(xl)])
        for i in range(1,3): 
            continue           
            y = xi_short(x,i) 
            assert y, lmap(hex, [x, xl, xi(xl,i), y])
            ref = to_short(xi(xl,i))
            assert y == ref, lmap(hex, [x, xl, xi(xl,i), ref, y])



#######################################################################
# test_gen_xi_ref
#######################################################################


@pytest.mark.gen_xi
@pytest.mark.parametrize("gen_xi_class, ref", [(gen, GenXi)])
def test_gen_xi_ref(gen_xi_class, ref, ntests=200):       
    xi, xi_ref = gen_xi_class.gen_xi_op_xi, ref.gen_xi_op_xi 

    x_old = 0
    for x in xi_samples(ntests):
        for i in range(0,3):
           res, ref_ = xi(x, i), xi_ref(x, i)
           assert res == ref_ , lmap(hex,[i, x,res, ref_ ])
        assert gen_xi_class.gen_xi_g_gray(x) == ref.gen_xi_g_gray(x)
        assert gen_xi_class.gen_xi_w2_gray(x) == ref.gen_xi_w2_gray(x)
        assert gen_xi_class.gen_xi_g_cocode(x) == ref.gen_xi_g_cocode(x)
        assert gen_xi_class.gen_xi_w2_cocode(x) == ref.gen_xi_w2_cocode(x)
        x_old = x
    xi_short = gen_xi_class.gen_xi_op_xi_short 
    to_leech = gen_xi_class.gen_xi_short_to_leech
    to_short = gen_xi_class.gen_xi_leech_to_short
    xi_short_ref = ref.gen_xi_op_xi_short 
    to_leech_ref = ref.gen_xi_short_to_leech
    for x in short_samples(ntests):
        xl = to_leech_ref(x)
        assert to_leech(x) == xl, lmap(hex,[x, xl,to_leech(x)])
        x_norm = normalize_short_vector_code(x)
        assert to_short(xl) == x_norm, lmap(hex,[xl, x, x_norm, to_short(xl)])
        for i in range(1,3):         
            ref = xi_short_ref(x,i)   
            assert xi_short(x,i) ==  ref, lmap(hex,[i, x, xi_short(x,i), ref])
            pass





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


"""Modelling an abstract group

Yet to be documented

"""
import sys
import re
import warnings
from copy import deepcopy
from random import sample, randint
from numbers import Integral, Number

from mmgroup.structures.parse_atoms import  eval_atom_expression        
from mmgroup.structures.parity import Parity



####################################################################
####################################################################
### Class AbstractGroup and helpers for that class
####################################################################
####################################################################

####################################################################
### Class AbstractGroupWord
####################################################################


class AbstractGroupWord(object):
    """Model an element of an abstract group.

    Users should not refer to this class directly. They should create 
    a group as an instance of subclass of class AbstractGroup and use 
    the methods of that group for creating elements.

    The standard group operations '*' '/' (=right multiplication with
    the inverse) and '**' are implemented here.

    g1 ** g2  means  g2**(-1) * g1 * g2  for group elements g1, g2.

    Here a group is an instance of (a subclass of) class AbstractGroup. 

    For each word a group 'g' should be passed as a keyword argument 
    'group = g'. If a subclass of class 'AbstractGroup' models one 
    group only, the corresponding subclass of this class may 
    contain a class attribute 'group' referring to that group. Then 
    the user  may contruct elements of that group using the 
    constructor of that subclass of this class.
    """
    def __init__(self, *args, **kwds):
        try:
            self.group = kwds['group']
        except:
            assert isinstance(self.group, AbstractGroup)

    # There is no need to modify any methods below this line.
    # You should overwrite the corresonding methods in the
    # subclasses of class AbstractGroup insead.
 
    def __eq__(self, other):    
        return( isinstance(other, AbstractGroupWord) 
            and  self.group == other.group
            and  self.group._equal_words(self, other)
        )

    def __ne__(self, other): 
        return not self.__eq__(other)

    def copy(self):
        """Return a deep copy of the group element"""
        return self.group.copy_word(self)

    def __imul__(self, other):
        """Implementation of the group multiplication"""
        g = self.group
        return g._imul(self, g._to_group(other))

    def __mul__(self, other):
        """Implementation of the group multiplication"""
        g = self.group
        try:
            return g._imul(g.copy_word(self), g._to_group(other))
        except (TypeError, NotImplementedError) as exc:
            try:
                myself = other.group._to_group(self)
                return myself.__imul__(other)
            except:
                raise exc
    def __rmul__(self, other):
        """Implementation of the reverse group multiplication"""
        g = self.group
        if isinstance(other, Parity):
            return other
        try:
            return g._imul(g._to_group(other), self)
        except (TypeError, NotImplementedError) as exc:
            try:
                myself = other.group._to_group(self)
                return other.__imul__(myself)
            except:
                raise exc

    def __itruediv__(self, other):
        """Implementation of the group division

        Here self / other    means    self * other**(-1) .
        """
        g = self.group
        return g._imul(self, g._invert(g._to_group(other)))

    def __truediv__(self, other):
        """Implementation of the group division

        Here self / other    means    self * other**(-1) .
        """
        g = self.group
        return g._imul(g.copy_word(self), g._invert(g._to_group(other)))

    def __rtruediv__(self, other):
        """Implementation of the reverse group division

        Here self / other    means    self * other**(-1) .
        """
        g = self.group
        return g._imul(g.copy_word(g._to_group(other)), g._invert(self))
      
    def __pow__(self, exp):
        """Implementation of the power operation

        This is exponentiation for integer eponents and conjugation
        if the exponent is a group element.
        """
        g = self.group
        if isinstance(exp, Integral):
            if exp > 0:
                res, start = g.copy_word(self), self
            elif exp == 0:
                return g.neutral()
            else:
                start, exp = g._invert(self), -exp
                res = g.copy_word(start) 
            for i in range(int(exp).bit_length() - 2, -1, -1):
                res = g._imul(res, res)
                if exp & (1 << i):
                    res = g._imul(res, start) 
            return res      
        elif isinstance(exp, AbstractGroupWord):
            e = self.group._to_group(exp) 
            return g._imul(g._imul(g._invert(e), self), e)
        elif isinstance(exp, Parity):
            one = self.group.neutral()
            if self * self == one:
                return self if other.value & 1 else one
            raise ValueError("Group element has not order 1 or 2")
        else:
            return NotImplemented

    def reduce(self, copy = False):
        """Reduce a group element

        We assume that the representation of a group element
        is not always given in a unique form.

        Assueme that ``g`` can be converted to a unique form (or  
        at least to a sufficiently simple form) that we will call 
        a **reduced** form. This method tries to achieve this goal. 

        This method may be applied to a group element without
        notice.

        If argument ``copy`` is True, a reduced copy of ``g``
        should be returned if ``g`` is not reduced.
        """
        return self.group.reduce(self, copy)

    def str(self):
        """Represent group element as a string"""
        try:
            return self.group.str_word(self)
        except NotImplementedError:
            return super(AbstractGroupWord, str)()
    __repr__ = str

    def as_tuples(self):
        """Convert group element to a list of tuples

        For a group element ``g`` the following should hold:

        ``g.group.word(*g.as_tuples()) == g`` .

        So passing the tuples in the list returned by this method
        as arguments to ``g.group`` or to ``g.group.word`` 
        reconstructs the element ``g``.

        This shows how to convert a group element to a list of tuples
        and vice versa.
        """
        return self.group.as_tuples(self)




####################################################################
### Class AbstractGroup
####################################################################


class AbstractGroup(object):
    """Model an abstract group"""
    word_type = AbstractGroupWord  # type of an element (=word) in the group

    def __init__(self, *data, **kwds):
        """Creating instances is only possible for concrete groups

         
        """
        pass

    ### The following methods must be overwritten ####################

    def __call__(self, *args):
        """Convert argumentss given by ``args``  to a group element
        """
        raise NotImplementedError




    def _imul(self, g1, g2):
        """Return product g1 * g2 of group elements g1 and g2.

        g1 may be destroyed but not g2.

        This method is called for elements g1 and g2 of the group
        'self' only. It should return the reduced product.
        """
        raise NotImplementedError("No multiplication in abstract group")


    def _invert(self, g1):
        """Return inverse g1**(-1) of group element g1.

        g1 must not be destroyed.

        This method is called for elements g1 of the group
        'self' only. It should return the reduced inverse.
        """
        raise NotImplementedError("No inversion in abstract group")
        
    ### The following methods should be overwritten ###################

    def copy_word(self, g1):
        """Return deep copy of group element ``g1``"""
        g_copy = deepcopy(g1)
        # Even a deep copy of an element is still in the same group
        g_copy.group = g1.group
        return g_copy

    def _equal_words(self, g1, g2):
        """Return True iff elements g1 and g2 are equal 

        This method is called for elements g1 and g2 of the group
        'self' only.
		
        In concrete group this method should be overwritten with
        a comparison of the relevant attributes of g1 and g2.

        Caution:
        Non-reduced words may be considered unequal even if they
        represent the same element. Use g1.reduce() or g1 * 1 to
        obtain the reduced form of g1. See method reduce() for
        details.
        """
        return g1 == g2


    def reduce(self, g, copy = False):
        """Reduce the word ``g`` which is an element of the group

        We assume that the representation of a group element
        is not always given in a unique form.

        Assueme that ``g`` can be converted to a unique form (or  
        at least to a sufficiently simple form) that we will call 
        a **reduced** form.  This method tries to achieve this goal. 

        This method may be applied to a group element without
        notice.

        If argument ``copy`` is True, a reduced copy of ``g``
        should be returned if ``g`` is not reduced.
        """
        return g




    def as_tuples(self, g):
        """Convert group element ``g`` to a list of tuples.

        The returned tuple should represent a reduced word.

        The sequence:: 

            l = g.group.as_tuples(g) 
            g1 = g.group(*l)

        should compute a group element ``g1`` with ``g1 == g``.
        """
        raise NotImplementedError("Abstract method")


    def str_word(self, g):
        """Convert group atom ``g`` to a string

        For an element ``g`` of this group ``g.group.str_word(g)``
        should be equivalent   to ``g.str()``.
        """
        raise NotImplementedError

                 
    ### The following methods need not be overwritten #################

    def neutral(self):
        """Return neutral element of the group"""
        return self.__call__()


  
    def _to_group(self, g):
        """Convert the object ``g`` to an element of this group

        This function performs an implicit conversion of the object 
        ``g`` to a group element. This function is applied in a 
        group operation.
        """
        if isinstance(g, AbstractGroupWord) and g.group == self:
            return g
        if g == 1:
            return self.neutral()
        err = "Cannot convert type '%s' object to group element"
        raise TypeError(err % type(g))
           


    ### The following methods should not be overwritten ###############




    def __contains__(self, other):
        """Return True iff 'other' is an element of the group"""
        try:
            if not isinstance(other, self.word_type): 
                 return False
            if other.group != self: 
                 return False
            return True
        except:
            return False





def singleton(cls):
    """Decorator to convert a class into a singleton

    This function should be used as a decorator for a class.

    This decorator turns a class is a singleton, so that all instances
    of that class are identical. Also, we do not allow any arguments 
    in the constructor of such a class.

    A typical use case is the decoration of a class derived from
    class ``AbstractGroup``. There are many different permutation
    groups, but (up to isomorphism) there in only one monster group.

    So if a class models the monster group, it makes sense to
    implement this class as a singleton.
    """
    cls.__instance = None # This will be the only instance of the class
    # Always return the same object
    def new_(cls, *args, **kwds):
        # Disable arguments in constructor
        if len(args) + len(kwds):
            ERR = "Arguments in constructor of a singleton are not allowed"
            raise TypeError(ERR)
        # Always return the same object
        if cls.__instance is None:
            cls.__instance =  object.__new__(cls) 
        return cls.__instance 
    cls.__new__ = staticmethod(new_)
    return cls



 